from __future__ import annotations

from typing import Optional

import attrs

from zyte_common_items.base import Item


@attrs.define(kw_only=True)
class Amenity(Item):
    """An amenity that a business place has"""

    #: Name of amenity.
    name: str

    #: Availability of the amenity.
    value: bool


@attrs.define(kw_only=True)
class ParentPlace(Item):
    """If the place is located inside another place, these are the details of the parent place."""

    #: Name of the parent place.
    name: str

    #: Identifier of the parent place.
    placeId: str


@attrs.define(kw_only=True)
class OpeningHoursItem(Item):
    """Specification of opening hours of a business place."""

    #: English weekday name.
    dayOfWeek: Optional[str] = None

    #: Opening time in ISO 8601 format, local time.
    opens: Optional[str] = None

    #: Closing time in ISO 8601 format, local time.
    closes: Optional[str] = None

    #: Day of the week, as it appears on the page, without processing.
    rawDayOfWeek: Optional[str] = None

    #: Opening time, as it appears on the page, without processing.
    rawOpens: Optional[str] = None

    #: Closing time, as it appears on the page, without processing.
    rawCloses: Optional[str] = None



"""
Components specific to :class:`~SocialMediaPost` item.
"""

from typing import Optional

import attrs

from zyte_common_items.base import Item


@attrs.define(kw_only=True)
class Reactions(Item):
    """Details of reactions to a post."""

    #: Number of times the post has been shared.
    reposts: Optional[int] = None

    #: Number of likes or other positive reactions to the post.
    likes: Optional[int] = None

    #: Number of dislikes or other negative reactions to the post.
    dislikes: Optional[int] = None


@attrs.define(kw_only=True)
class SocialMediaPostAuthor(Item):
    """Details of the author of a social media post."""

    #: The number of the followers that observe the author.
    numberOfFollowers: Optional[int] = None

    #: The number of the users that the author follows.
    numberOfFollowing: Optional[int] = None

    #: The date of the creation of the author's account.
    dateAccountCreated: Optional[str] = None

    #: The location of the author, if it's available in the author profile. Country or city location only.
    location: Optional[str] = None

    #: Indication if the author's account is verified.
    isVerified: Optional[bool] = None



from web_poet import Extractor, SelectorExtractor

from .items import ProductFromList, ProductVariant
from .processors import gtin_processor, price_processor, simple_price_processor


class _ProductProcessors:
    price = [price_processor]
    regularPrice = [simple_price_processor]


class ProductFromListExtractor(Extractor[ProductFromList]):
    """:class:`~web_poet.pages.Extractor` for :class:`ProductFromList`."""

    class Processors(_ProductProcessors):
        pass


class ProductFromListSelectorExtractor(SelectorExtractor[ProductFromList]):
    """:class:`~web_poet.pages.SelectorExtractor` for
    :class:`ProductFromList`."""

    class Processors(_ProductProcessors):
        pass


class _ProductVariantProcessors(_ProductProcessors):
    gtin = [gtin_processor]


class ProductVariantExtractor(Extractor[ProductVariant]):
    """:class:`~web_poet.pages.Extractor` for :class:`ProductVariant`."""

    class Processors(_ProductVariantProcessors):
        pass


class ProductVariantSelectorExtractor(SelectorExtractor[ProductVariant]):
    """:class:`~web_poet.pages.SelectorExtractor` for
    :class:`ProductVariant`."""

    class Processors(_ProductVariantProcessors):
        pass



from typing import List, Optional

import attrs
from web_poet import Returns

from zyte_common_items.components import ProbabilityRequest, Request
from zyte_common_items.fields import auto_field
from zyte_common_items.items import ArticleNavigation, ArticleNavigationMetadata

from .base import BasePage, Page
from .mixins import HasMetadata


class BaseArticleNavigationPage(
    BasePage, Returns[ArticleNavigation], HasMetadata[ArticleNavigationMetadata]
):
    """:class:`BasePage` subclass for :class:`ArticleNavigation`."""


class ArticleNavigationPage(
    Page, Returns[ArticleNavigation], HasMetadata[ArticleNavigationMetadata]
):
    """:class:`Page` subclass for :class:`ArticleNavigation`."""


@attrs.define
class AutoArticleNavigationPage(BaseArticleNavigationPage):
    article_navigation: ArticleNavigation

    @auto_field
    def categoryName(self) -> Optional[str]:
        return self.article_navigation.categoryName

    @auto_field
    def items(self) -> Optional[List[ProbabilityRequest]]:
        return self.article_navigation.items

    @auto_field
    def metadata(self) -> Optional[ArticleNavigationMetadata]:
        return self.article_navigation.metadata

    @auto_field
    def nextPage(self) -> Optional[Request]:
        return self.article_navigation.nextPage

    @auto_field
    def pageNumber(self) -> Optional[int]:
        return self.article_navigation.pageNumber

    @auto_field
    def subCategories(self) -> Optional[List[ProbabilityRequest]]:
        return self.article_navigation.subCategories

    @auto_field
    def url(self) -> Optional[str]:
        return self.article_navigation.url



from typing import List, Optional

import attrs
from web_poet import Returns

from zyte_common_items.components import (
    AdditionalProperty,
    AggregateRating,
    Brand,
    Breadcrumb,
    Gtin,
    Image,
)
from zyte_common_items.fields import auto_field
from zyte_common_items.items import Product, ProductMetadata, ProductVariant
from zyte_common_items.processors import (
    brand_processor,
    breadcrumbs_processor,
    description_html_processor,
    description_processor,
    gtin_processor,
    images_processor,
    price_processor,
    rating_processor,
    simple_price_processor,
)

from .base import BasePage, Page
from .mixins import DescriptionMixin, HasMetadata, PriceMixin


class BaseProductPage(
    BasePage,
    DescriptionMixin,
    PriceMixin,
    Returns[Product],
    HasMetadata[ProductMetadata],
):
    """:class:`BasePage` subclass for :class:`Product`."""

    class Processors(BasePage.Processors):
        aggregateRating = [rating_processor]
        brand = [brand_processor]
        breadcrumbs = [breadcrumbs_processor]
        description = [description_processor]
        descriptionHtml = [description_html_processor]
        gtin = [gtin_processor]
        price = [price_processor]
        regularPrice = [simple_price_processor]
        images = [images_processor]


class ProductPage(
    Page, DescriptionMixin, PriceMixin, Returns[Product], HasMetadata[ProductMetadata]
):
    """:class:`Page` subclass for :class:`Product`."""

    class Processors(Page.Processors):
        aggregateRating = [rating_processor]
        brand = [brand_processor]
        breadcrumbs = [breadcrumbs_processor]
        description = [description_processor]
        descriptionHtml = [description_html_processor]
        gtin = [gtin_processor]
        price = [price_processor]
        regularPrice = [simple_price_processor]
        images = [images_processor]


@attrs.define
class AutoProductPage(BaseProductPage):
    product: Product

    @auto_field
    def additionalProperties(self) -> Optional[List[AdditionalProperty]]:
        return self.product.additionalProperties

    @auto_field
    def aggregateRating(self) -> Optional[AggregateRating]:
        return self.product.aggregateRating

    @auto_field
    def availability(self) -> Optional[str]:
        return self.product.availability

    @auto_field
    def brand(self) -> Optional[Brand]:
        return self.product.brand

    @auto_field
    def breadcrumbs(self) -> Optional[List[Breadcrumb]]:
        return self.product.breadcrumbs

    @auto_field
    def canonicalUrl(self) -> Optional[str]:
        return self.product.canonicalUrl

    @auto_field
    def color(self) -> Optional[str]:
        return self.product.color

    @auto_field
    def currency(self) -> Optional[str]:
        return self.product.currency

    @auto_field
    def currencyRaw(self) -> Optional[str]:
        return self.product.currencyRaw

    @auto_field
    def description(self) -> Optional[str]:
        return self.product.description

    @auto_field
    def descriptionHtml(self) -> Optional[str]:
        return self.product.descriptionHtml

    @auto_field
    def features(self) -> Optional[List[str]]:
        return self.product.features

    @auto_field
    def gtin(self) -> Optional[List[Gtin]]:
        return self.product.gtin

    @auto_field
    def images(self) -> Optional[List[Image]]:
        return self.product.images

    @auto_field
    def mainImage(self) -> Optional[Image]:
        return self.product.mainImage

    @auto_field
    def metadata(self) -> Optional[ProductMetadata]:
        return self.product.metadata

    @auto_field
    def mpn(self) -> Optional[str]:
        return self.product.mpn

    @auto_field
    def name(self) -> Optional[str]:
        return self.product.name

    @auto_field
    def price(self) -> Optional[str]:
        return self.product.price

    @auto_field
    def productId(self) -> Optional[str]:
        return self.product.productId

    @auto_field
    def regularPrice(self) -> Optional[str]:
        return self.product.regularPrice

    @auto_field
    def size(self) -> Optional[str]:
        return self.product.size

    @auto_field
    def sku(self) -> Optional[str]:
        return self.product.sku

    @auto_field
    def style(self) -> Optional[str]:
        return self.product.style

    @auto_field
    def url(self) -> str:
        return self.product.url

    @auto_field
    def variants(self) -> Optional[List[ProductVariant]]:
        return self.product.variants



from web_poet import ItemPage

from zyte_common_items.items import SearchRequestTemplate, SearchRequestTemplateMetadata

from .mixins import HasMetadata


class SearchRequestTemplatePage(
    ItemPage[SearchRequestTemplate], HasMetadata[SearchRequestTemplateMetadata]
):
    pass



import warnings
from typing import List

from .request import ProbabilityRequest, Request


def request_list_processor(request_list: List[Request]) -> List[ProbabilityRequest]:
    """Deprecated. Please use
    :func:`zyte_common_items.processors.probability_request_list_processor` instead."""
    warnings.warn(
        message="request_list_processor is deprecated in favor of "
        "zyte_common_items.processors.probability_request_list_processor",
        category=DeprecationWarning,
        stacklevel=2,
    )
    from zyte_common_items.processors import probability_request_list_processor

    return probability_request_list_processor(request_list)



import warnings
from typing import Optional
from warnings import catch_warnings

import attrs
import pytest
from web_poet import HttpResponse, RequestUrl, ResponseUrl, Returns, field
from web_poet.fields import get_fields_dict

import zyte_common_items
from zyte_common_items import (
    BaseProductListPage,
    BaseProductNavigationPage,
    BaseProductPage,
    HasMetadata,
    Metadata,
    Page,
    ProbabilityMetadata,
    ProbabilityRequest,
    Product,
    ProductFromList,
    ProductListMetadata,
    ProductListPage,
    ProductMetadata,
    ProductPage,
    Request,
)
from zyte_common_items._dateutils import utcnow
from zyte_common_items.fields import is_auto_field


@pytest.mark.parametrize(
    "page_class",
    (
        BaseProductPage,
        BaseProductListPage,
    ),
)
def test_base_pages_default(page_class):
    page = page_class(request_url=RequestUrl("https://example.com"))
    assert page.url == "https://example.com"
    assert isinstance(page.url, str)
    # NOTE: test_metadata covers .metadata


@pytest.mark.parametrize(
    "page_class",
    (
        ProductPage,
        ProductListPage,
    ),
)
def test_pages_default(page_class):
    url = ResponseUrl("https://example.com")
    html = b"""
    <!DOCTYPE html>
    <html>
        <body>
            <h1>Foo</h1>
        </body>
    </html>
    """
    response = HttpResponse(url=url, body=html)
    page = page_class(response=response)
    assert page.url == "https://example.com"
    assert isinstance(page.url, str)
    # NOTE: test_metadata covers .metadata


@pytest.mark.asyncio
async def test_example():
    datetime_before = utcnow().replace(microsecond=0)

    @attrs.define
    class BookPage(ProductPage):
        @field
        def name(self):
            return self.css("h1::text").get()

    url = ResponseUrl("https://example.com/books/1")
    html = b"""
    <!DOCTYPE html>
    <html>
        <body>
            <h1>Foo</h1>
        </body>
    </html>
    """
    response = HttpResponse(url=url, body=html)

    item = await BookPage(response=response).to_item()

    assert item.url == str(url)
    assert item.name == "Foo"
    assert item.metadata.probability == 1.0

    item_datetime = item.metadata.get_date_downloaded_parsed()
    datetime_after = utcnow().replace(microsecond=0)
    assert datetime_before <= item_datetime <= datetime_after


@pytest.mark.asyncio
async def test_mixin_leak():
    """https://github.com/zytedata/zyte-common-items/pull/29"""

    class MyProductListPage(ProductListPage):
        @field
        def products(self):
            return [{"name": "foo"}, {"name": "bar"}]

    class MyProductPage(ProductPage):
        @field
        def brand(self):
            return "baz"

    assert set(get_fields_dict(MyProductListPage)) == {"metadata", "products", "url"}
    assert set(get_fields_dict(MyProductPage)) == {
        "brand",
        "currency",
        "currencyRaw",
        "description",
        "descriptionHtml",
        "metadata",
        "url",
    }


@pytest.mark.asyncio
async def test_no_item_found_Page():
    class MyProductPage(ProductPage):
        def validate_input(self):
            return self.no_item_found()

    response = HttpResponse(url="http://example.com", body=b"<html></html>")
    page = MyProductPage(response=response)

    item = await page.to_item()
    assert item.metadata.probability == 0
    assert item.url == "http://example.com"


@pytest.mark.asyncio
async def test_no_item_found_BasePage():
    class MyProductPage(BaseProductPage):
        def validate_input(self):
            return self.no_item_found()

    page = MyProductPage(request_url=RequestUrl("http://example.com"))

    item = await page.to_item()
    assert item.metadata.probability == 0
    assert item.url == "http://example.com"


def test_page_pairs():
    """For every page, a base page and an auto page, and vice versa."""
    pages = {
        obj_name
        for obj_name in zyte_common_items.__dict__
        if (
            not (
                obj_name.startswith("Base")
                or obj_name.startswith("Auto")
                or obj_name.endswith("RequestTemplatePage")
            )
            and obj_name.endswith("Page")
            and obj_name != "Page"
        )
    }

    actual_base_pages = {
        obj_name
        for obj_name in zyte_common_items.__dict__
        if (
            obj_name.startswith("Base")
            and obj_name.endswith("Page")
            and obj_name != "BasePage"
        )
    }
    expected_base_pages = {f"Base{page}" for page in pages}
    assert actual_base_pages == expected_base_pages

    actual_auto_pages = {
        obj_name
        for obj_name in zyte_common_items.__dict__
        if (obj_name.startswith("Auto") and obj_name.endswith("Page"))
    }
    expected_auto_pages = {f"Auto{page}" for page in pages}
    assert actual_auto_pages == expected_auto_pages


def test_matching_items():
    """For every page, an item."""
    pages = {
        obj_name
        for obj_name in zyte_common_items.__dict__
        if (
            not (obj_name.startswith("Base") or obj_name.startswith("Auto"))
            and obj_name.endswith("Page")
            and obj_name != "Page"
        )
    }
    for page in pages:
        item = page[:-4]
        assert item in zyte_common_items.__dict__


METADATA_FIELDS = {
    "Article": {"dateDownloaded", "probability", "validationMessages"},
    "ArticleList": {"dateDownloaded", "validationMessages"},
    "ArticleNavigation": {"dateDownloaded", "validationMessages"},
    "BusinessPlace": {
        "dateDownloaded",
        "probability",
        "searchText",
        "validationMessages",
    },
    "Product": {"dateDownloaded", "probability", "validationMessages"},
    "ProductList": {"dateDownloaded", "validationMessages"},
    "ProductNavigation": {"dateDownloaded", "validationMessages"},
    "RealEstate": {"dateDownloaded", "probability", "validationMessages"},
    "JobPosting": {"dateDownloaded", "probability", "searchText", "validationMessages"},
    "SocialMediaPost": {
        "dateDownloaded",
        "probability",
        "searchText",
        "validationMessages",
    },
}


def check_default_metadata(cls, kwargs, item_name):
    start = utcnow().replace(microsecond=0)

    obj = cls(**kwargs)

    metadata_cls = zyte_common_items.__dict__[f"{item_name}Metadata"]
    assert type(obj.metadata) is metadata_cls

    expected_fields = METADATA_FIELDS[item_name]

    def allow_field(field_name):
        for prefix in ["_", "from_", "get_"]:
            if field_name.startswith(prefix):
                return False
        if field_name == "cast":
            return False
        return True

    actual_fields = {field for field in dir(obj.metadata) if allow_field(field)}
    error_message = (
        f"{metadata_cls}: actual fields ({actual_fields}) != expected fields "
        f"({expected_fields})"
    )
    assert actual_fields == expected_fields, error_message

    if "dateDownloaded" in actual_fields:
        assert isinstance(
            obj.metadata.dateDownloaded, str
        ), f"{cls} does not get dateDownloaded set by default"
        assert obj.metadata.dateDownloaded.endswith("Z")
        actual = obj.metadata.get_date_downloaded_parsed()
        end = utcnow().replace(microsecond=0)
        assert start <= actual <= end

    if "probability" in actual_fields:
        assert isinstance(obj.metadata.probability, float)
        assert obj.metadata.probability == 1.0


def test_metadata():
    """Test metadata expectations for pages.

    For every type-specific page and base page:

    -   There must be a matching metadata class.

    -   The metadata attribute must be of that metadata class.

    -   The default value of metadata must have dateDownloaded and probability
        filled if available in the metadata class.
    """
    pages = {
        obj_name
        for obj_name in zyte_common_items.__dict__
        if (
            not (
                obj_name.startswith("Base")
                or obj_name.startswith("Auto")
                or obj_name.endswith("RequestTemplatePage")
            )
            and obj_name.endswith("Page")
            and obj_name != "Page"
        )
    }

    for page in pages:
        item_name = page[:-4]

        page_cls = getattr(zyte_common_items, page)
        response_url = ResponseUrl("https://example.com")
        html = b"""
        <!DOCTYPE html>
        <html>
            <body>
                <h1>Foo</h1>
            </body>
        </html>
        """
        page_kwargs = {"response": HttpResponse(url=response_url, body=html)}
        check_default_metadata(page_cls, page_kwargs, item_name)

        base_page = f"Base{page}"
        base_page_cls = getattr(zyte_common_items, base_page)
        base_page_kwargs = {"request_url": RequestUrl("https://example.com")}
        check_default_metadata(base_page_cls, base_page_kwargs, item_name)


def test_metadata_generic():
    """Ensure that when a subclass returns an instance of the generic Metadata
    class, it gets replaced by the corresponding, more specific metadata
    class."""

    class CustomProductPage(ProductPage):
        @field
        def metadata(self):
            return Metadata(dateDownloaded="foo", probability=0.5)

    url = ResponseUrl("https://example.com")
    html = b"""
    <!DOCTYPE html>
    <html>
        <body>
            <h1>Foo</h1>
        </body>
    </html>
    """
    page1 = CustomProductPage(response=HttpResponse(url=url, body=html))
    with catch_warnings():
        warnings.simplefilter("error")
        metadata1 = page1.metadata
    assert type(metadata1) is ProductMetadata
    assert metadata1.dateDownloaded == "foo"
    assert metadata1.probability == 0.5

    class ExtraAttrProductPage(ProductPage):
        @field
        def metadata(self):
            return Metadata(dateDownloaded="foo", probability=0.5, searchText="bar")

    page2 = ExtraAttrProductPage(response=HttpResponse(url=url, body=html))
    with pytest.warns(RuntimeWarning, match=r"dropping the non-default values"):
        metadata2 = page2.metadata
    assert type(metadata2) is ProductMetadata
    assert metadata2.dateDownloaded == "foo"
    assert metadata2.probability == 0.5

    class DefaultAttrProductListPage(ProductListPage):
        @field
        def metadata(self):
            return Metadata(dateDownloaded="foo", probability=1.0)

        @field
        def products(self):
            return [ProductFromList(url=url, metadata=Metadata())]

    page3 = DefaultAttrProductListPage(response=HttpResponse(url=url, body=html))
    with catch_warnings():
        warnings.simplefilter("error")
        metadata3 = page3.metadata
        metadata4 = page3.products[0].metadata
    assert type(metadata3) is ProductListMetadata
    assert metadata3.dateDownloaded == "foo"
    assert not hasattr(metadata3, "probability")
    assert type(metadata4) is ProbabilityMetadata
    assert metadata4.probability == 1.0


def test_metadata_override():
    """Test using HasMetadata to override the metadata class."""

    @attrs.define(kw_only=True)
    class CustomProductMetadata(ProductMetadata):
        new_field: Optional[str] = None

    class CustomProductPage(ProductPage, HasMetadata[CustomProductMetadata]):
        @field
        def metadata(self):
            return Metadata(dateDownloaded="foo", probability=0.5)

    url = ResponseUrl("https://example.com")
    html = b"""
    <!DOCTYPE html>
    <html>
        <body>
            <h1>Foo</h1>
        </body>
    </html>
    """
    page = CustomProductPage(response=HttpResponse(url=url, body=html))
    metadata = page.metadata
    assert type(metadata) is CustomProductMetadata
    assert metadata.new_field is None


def test_hasmetadata_inheritance():
    """Ensure that a subclass with just Returns doesn't break _get_metadata_class()."""

    class MyProduct(Product):
        pass

    class CustomProductPage(ProductPage, Returns[MyProduct]):
        pass

    url = ResponseUrl("https://example.com")
    html = b"""
    <!DOCTYPE html>
    <html>
        <body>
            <h1>Foo</h1>
        </body>
    </html>
    """
    page = CustomProductPage(response=HttpResponse(url=url, body=html))
    metadata = page.metadata
    assert type(metadata) is ProductMetadata


def test_metadata_cls_none():
    """Ensure that metadata_cls can be None."""

    class CustomProductPage(Page, Returns[Product]):
        pass

    url = ResponseUrl("https://example.com")
    html = b"""
    <!DOCTYPE html>
    <html>
        <body>
            <h1>Foo</h1>
        </body>
    </html>
    """
    page = CustomProductPage(response=HttpResponse(url=url, body=html))
    with pytest.raises(ValueError):
        page.metadata


def test_request():
    """Test Request class conversion"""

    class CustomNavPage(BaseProductNavigationPage):
        @field
        def subCategories(self):
            return [Request(url="https://example.com")]

        @field
        def items(self):
            return [Request(url="https://example.com")]

        @field
        def nextPage(self):
            return Request(url="https://example.com")

    url = RequestUrl("https://example.com")
    page = CustomNavPage(url)
    for request in [page.subCategories[0], page.items[0]]:
        assert type(request) is ProbabilityRequest
        assert request.metadata is not None
        assert request.metadata.probability == 1.0
    assert type(page.nextPage) is Request


def test_auto_fields():
    """Every field of an Auto-prefixed class should have ``auto_field`` set to
    ``True`` in its field metadata."""
    auto_page_names = {
        obj_name
        for obj_name in zyte_common_items.__dict__
        if (obj_name.startswith("Auto") and obj_name.endswith("Page"))
    }
    for auto_page_name in auto_page_names:
        auto_page_cls = zyte_common_items.__dict__[auto_page_name]
        for field_name in get_fields_dict(auto_page_cls):
            assert is_auto_field(auto_page_cls, field_name)



from typing import List, Optional

import attrs

from zyte_common_items.base import Item
from zyte_common_items.components import ListMetadata, ProbabilityRequest, Request
from zyte_common_items.converters import (
    to_metadata_optional,
    to_probability_request_list_optional,
    url_to_str,
)


@attrs.define(kw_only=True)
class ArticleNavigationMetadata(ListMetadata):
    """Metadata class for :data:`zyte_common_items.ArticleNavigation.metadata`."""


@attrs.define(kw_only=True)
class ArticleNavigation(Item):
    """Represents the navigational aspects of an article listing webpage.

    See :class:`ArticleList`.
    """

    #: Main URL from which the data is extracted.
    url: str = attrs.field(converter=url_to_str)

    #: Name of the category/page.
    #:
    #: Format:
    #:
    #: - trimmed (no whitespace at the beginning or the end of the description string)
    categoryName: Optional[str] = None

    #: List of sub-category links ordered by their position in the page.
    subCategories: Optional[List[ProbabilityRequest]] = attrs.field(
        default=None, converter=to_probability_request_list_optional, kw_only=True  # type: ignore[misc]
    )

    #: Links to listed items in order of appearance.
    items: Optional[List[ProbabilityRequest]] = attrs.field(
        default=None, converter=to_probability_request_list_optional, kw_only=True  # type: ignore[misc]
    )

    #: A link to the next page, if available.
    nextPage: Optional[Request] = None

    #: Number of the current page.
    #:
    #: It should only be extracted if the webpage shows a page number.
    #:
    #: It must be 1-based. For example, if the first page of a listing is
    #: numbered as 0 on the website, it should be extracted as `1` nonetheless.
    pageNumber: Optional[int] = None

    #: Data extraction process metadata.
    metadata: Optional[ArticleNavigationMetadata] = attrs.field(
        default=None, converter=to_metadata_optional(ArticleNavigationMetadata), kw_only=True  # type: ignore[misc]
    )



"""The ``Item`` class should be used as the parent class for data containers."""

from collections import ChainMap
from typing import Dict, List, Optional, Union, get_args, get_origin, get_type_hints

import attrs

from .util import split_in_unknown_and_known_fields

_Trail = Optional[str]
_UNDEFINED = object()


def is_data_container(cls_or_obj):
    """Used for discerning classes/instances if they are part of the Zyte Common
    Item definitions.
    """
    try:
        return issubclass(cls_or_obj, Item)
    except TypeError:
        # must be an instance rather than a class
        return isinstance(cls_or_obj, Item)


class _ItemBase:
    # Reserving a slot for _unknown_fields_dict.
    # This is done in a base class because otherwise attr.s won't pick it up
    __slots__ = ("_unknown_fields_dict",)


def _get_import_path(obj: type):
    return f"{obj.__module__}.{obj.__qualname__}"


def _extend_trail(trail: _Trail, key: Union[int, str]):
    if isinstance(key, str):
        if not trail:
            trail = key
        else:
            trail += f".{key}"
    else:
        assert isinstance(key, int)
        item = f"[{key}]"
        if not trail:
            trail = item
        else:
            trail += item
    return trail


@attrs.define
class ProbabilityMixin:
    """Provides :meth:`get_probability` to make it easier to access the
    probability of an item or item component that is nested under its metadata
    attribute."""

    def get_probability(self) -> Optional[float]:
        """Returns the item probability if available, otherwise ``None``."""
        if metadata := getattr(self, "metadata", None):
            return getattr(metadata, "probability", None)
        return None


@attrs.define
class Item(ProbabilityMixin, _ItemBase):
    """Base class for :ref:`items <items>`."""

    def __attrs_post_init__(self):
        self._unknown_fields_dict = {}  # type: ignore[misc]

    @classmethod
    def from_dict(cls, item: Optional[Dict]):
        """Read an item from a dictionary."""
        return cls._from_dict(item)

    @classmethod
    def _from_dict(cls, item: Optional[Dict], *, trail: _Trail = None):
        """Read an item from a dictionary."""
        if item is None:
            return None

        if not isinstance(item, dict):
            path = _get_import_path(cls)
            if not trail:
                prefix = "Expected"
            else:
                prefix = f"Expected {trail} to be"
            raise ValueError(f"{prefix} a dict with fields from {path}, got {item!r}.")

        item = cls._apply_field_types_to_sub_fields(item, trail=trail)
        unknown_fields, known_fields = split_in_unknown_and_known_fields(item, cls)
        obj = cls(**known_fields)  # type: ignore
        obj._unknown_fields_dict = unknown_fields  # type: ignore[misc]
        return obj

    @classmethod
    def from_list(cls, items: Optional[List[Dict]], *, trail: _Trail = None) -> List:
        """Read items from a list."""
        return cls._from_list(items)

    @classmethod
    def _from_list(cls, items: Optional[List[Dict]], *, trail: _Trail = None) -> List:
        """Read items from a list."""
        result = []
        for index, item in enumerate(items or []):
            index_trail = _extend_trail(trail, index)
            result.append(cls._from_dict(item, trail=index_trail))
        return result

    @classmethod
    def _apply_field_types_to_sub_fields(cls, item: Dict, trail: _Trail = None):
        """This applies the correct data container class for some of the fields
        that need them.

        Specifically, this traverses recursively each field to determine the proper
        data container class based on the type annotations. This could handle both
        ``list`` and ``object`` type requirements. For example:

            * Article having ``breadcrumbs: List[Breadcrumb]``
            * Product having ``brand: Optional[Brand]``

        Moreover, fields that are not defined to be part of data container
        classes will be ignored. For example:

            * Article having ``headline: Optional[str]``
            * Product having ``name: Optional[str]``
        """
        from_dict, from_list = {}, {}

        annotations = ChainMap(*(get_type_hints(c) for c in cls.__mro__))
        for field, type_annotation in annotations.items():
            origin = get_origin(type_annotation)
            is_optional = False
            if origin == Union:
                field_classes = get_args(type_annotation)
                if len(field_classes) != 2 or not isinstance(None, field_classes[1]):
                    path = f"{_get_import_path(cls)}.{field}"
                    raise ValueError(
                        f"{path} is annotated with {type_annotation}. Fields "
                        f"should only be annotated with one type (or "
                        f"optional)."
                    )
                is_optional = len(field_classes) == 2 and isinstance(
                    None, field_classes[1]
                )
                type_annotation = field_classes[0]
                origin = get_origin(type_annotation)

            if origin is list:
                value = item.get(field, _UNDEFINED)
                if (
                    not isinstance(value, list)
                    and value is not _UNDEFINED
                    and not (is_optional and value is None)
                ):
                    field_trail = _extend_trail(trail, field)
                    raise ValueError(
                        f"Expected {field_trail} to be a list, got " f"{value!r}."
                    )
                type_annotation = get_args(type_annotation)[0]
                if is_data_container(type_annotation):
                    from_list[field] = type_annotation
            elif is_data_container(type_annotation):
                from_dict[field] = type_annotation

        if from_dict or from_list:
            item = dict(**item)
            for key, cls in (from_dict or {}).items():
                key_trail = _extend_trail(trail, key)
                value = item.get(key)
                if value is not None and not isinstance(value, dict):
                    path = _get_import_path(cls)
                    raise ValueError(
                        f"Expected {key_trail} to be a dict with fields "
                        f"from {path}, got {value!r}."
                    )
                item[key] = cls._from_dict(value, trail=key_trail)
            for key, cls in (from_list or {}).items():
                key_trail = _extend_trail(trail, key)
                value = item.get(key)
                if value is not None and not isinstance(value, list):
                    path = _get_import_path(cls)
                    raise ValueError(
                        f"Expected {key_trail} to be a list of dicts "
                        f"with fields from {path}, got {value!r}."
                    )
                item[key] = cls._from_list(value, trail=key_trail)

        return item



# flake8: noqa
from ._compat import RequestListCaster
from .article import Article, ArticleMetadata
from .article_list import ArticleFromList, ArticleList, ArticleListMetadata
from .article_navigation import ArticleNavigation, ArticleNavigationMetadata
from .business_place import BusinessPlace, BusinessPlaceMetadata
from .job_posting import JobPosting, JobPostingMetadata
from .product import Product, ProductMetadata, ProductVariant
from .product_list import ProductFromList, ProductList, ProductListMetadata
from .product_navigation import ProductNavigation, ProductNavigationMetadata
from .real_estate import RealEstate, RealEstateMetadata
from .search_request_template import (
    SearchRequestTemplate,
    SearchRequestTemplateMetadata,
)
from .social_media_post import SocialMediaPost, SocialMediaPostMetadata



from __future__ import annotations

from typing import List, Optional

import attrs

from zyte_common_items.base import Item
from zyte_common_items.components import (
    Author,
    Breadcrumb,
    Image,
    ListMetadata,
    ProbabilityMetadata,
)
from zyte_common_items.converters import (
    to_metadata_optional,
    url_to_str,
    url_to_str_optional,
)


@attrs.define(kw_only=True)
class ArticleListMetadata(ListMetadata):
    """Metadata class for :data:`zyte_common_items.ArticleList.metadata`."""


@attrs.define(slots=True, kw_only=True)
class ArticleList(Item):
    """Article list from an article listing page.

    The :attr:`url` attribute is the only required attribute, all other fields
    are optional.
    """

    #: The main URL of the article list.
    #:
    #: The URL of the final response, after any redirects.
    #:
    #: Required attribute.
    #:
    #: In case there is no article list data on the page or the page was not
    #: reached, the returned item still contain this URL field and all the
    #: other available datapoints.
    url: str = attrs.field(converter=url_to_str)

    #: Canonical form of the URL, as indicated by the website.
    #:
    #: See also ``url``.
    canonicalUrl: Optional[str] = attrs.field(
        default=None, converter=url_to_str_optional, kw_only=True
    )

    #: List of article details found on the page.
    #:
    #: The order of the articles reflects their position on the page.
    articles: Optional[List[ArticleFromList]] = None

    #: Webpage `breadcrumb trail`_.
    #:
    #: .. _Breadcrumb trail: https://en.wikipedia.org/wiki/Breadcrumb_navigation
    breadcrumbs: Optional[List[Breadcrumb]] = None

    #: Data extraction process metadata.
    metadata: Optional[ArticleListMetadata] = attrs.field(
        default=None, converter=to_metadata_optional(ArticleListMetadata), kw_only=True  # type: ignore[misc]
    )


@attrs.define(slots=True, kw_only=True)
class ArticleFromList(Item):
    """Article from an article list from an article listing page.

    See :class:`ArticleList`.
    """

    #: Clean text of the article, including sub-headings, with newline
    #: separators.
    #:
    #: Format:
    #:
    #: - trimmed (no whitespace at the beginning or the end of the body
    #:   string),
    #: - line breaks included,
    #: - no length limit,
    #: - no normalization of Unicode characters.
    articleBody: Optional[str] = None

    #: All authors of the article.
    authors: Optional[List[Author]] = None

    #: Publication date of the article.
    #:
    #: Format: ISO 8601 format: "YYYY-MM-DDThh:mm:ssZ" or
    #: "YYYY-MM-DDThh:mm:ss±zz:zz".
    #:
    #: With timezone, if available.
    #:
    #: If the actual publication date is not found, the date of the last
    #: modification is used instead.
    datePublished: Optional[str] = None

    #: Same date as
    #: :attr:`~zyte_common_items.ArticleFromList.datePublished`, but
    #: :before parsing/normalization, i.e. as it appears on the website.
    datePublishedRaw: Optional[str] = None

    #: Headline or title.
    headline: Optional[str] = None

    #: Language of the article, as an ISO 639-1 language code.
    #:
    #: Sometimes the article language is not the same as the web page overall
    #: language.
    inLanguage: Optional[str] = None

    #: Main image.
    mainImage: Optional[Image] = None

    #: All images.
    images: Optional[List[Image]] = None

    #: Data extraction process metadata.
    metadata: Optional[ProbabilityMetadata] = attrs.field(
        default=None,
        converter=to_metadata_optional(ProbabilityMetadata),  # type: ignore[misc]
        kw_only=True,
    )

    #: Main URL.
    url: Optional[str] = attrs.field(
        default=None, converter=url_to_str_optional, kw_only=True
    )



from typing import List, Optional

import attrs
from web_poet import Returns

from zyte_common_items.components import (
    AdditionalProperty,
    Address,
    Breadcrumb,
    Image,
    RealEstateArea,
)
from zyte_common_items.fields import auto_field
from zyte_common_items.items import RealEstate, RealEstateMetadata

from ..processors import breadcrumbs_processor, description_processor, images_processor
from .base import BasePage, Page
from .mixins import HasMetadata


class BaseRealEstatePage(
    BasePage, Returns[RealEstate], HasMetadata[RealEstateMetadata]
):
    """:class:`BasePage` subclass for :class:`RealEstate`."""

    class Processors(BasePage.Processors):
        breadcrumbs = [breadcrumbs_processor]
        description = [description_processor]
        images = [images_processor]


class RealEstatePage(Page, Returns[RealEstate], HasMetadata[RealEstateMetadata]):
    """:class:`Page` subclass for :class:`RealEstate`."""

    class Processors(Page.Processors):
        breadcrumbs = [breadcrumbs_processor]
        description = [description_processor]
        images = [images_processor]


@attrs.define
class AutoRealEstatePage(BaseRealEstatePage):
    real_estate: RealEstate

    @auto_field
    def additionalProperties(self) -> Optional[List[AdditionalProperty]]:
        return self.real_estate.additionalProperties

    @auto_field
    def address(self) -> Optional[Address]:
        return self.real_estate.address

    @auto_field
    def area(self) -> Optional[RealEstateArea]:
        return self.real_estate.area

    @auto_field
    def breadcrumbs(self) -> Optional[List[Breadcrumb]]:
        return self.real_estate.breadcrumbs

    @auto_field
    def currency(self) -> Optional[str]:
        return self.real_estate.currency

    @auto_field
    def currencyRaw(self) -> Optional[str]:
        return self.real_estate.currencyRaw

    @auto_field
    def datePublished(self) -> Optional[str]:
        return self.real_estate.datePublished

    @auto_field
    def datePublishedRaw(self) -> Optional[str]:
        return self.real_estate.datePublishedRaw

    @auto_field
    def description(self) -> Optional[str]:
        return self.real_estate.description

    @auto_field
    def images(self) -> Optional[List[Image]]:
        return self.real_estate.images

    @auto_field
    def mainImage(self) -> Optional[Image]:
        return self.real_estate.mainImage

    @auto_field
    def metadata(self) -> Optional[RealEstateMetadata]:
        return self.real_estate.metadata

    @auto_field
    def name(self) -> Optional[str]:
        return self.real_estate.name

    @auto_field
    def numberOfBathroomsTotal(self) -> Optional[int]:
        return self.real_estate.numberOfBathroomsTotal

    @auto_field
    def numberOfBedrooms(self) -> Optional[int]:
        return self.real_estate.numberOfBedrooms

    @auto_field
    def numberOfFullBathrooms(self) -> Optional[int]:
        return self.real_estate.numberOfFullBathrooms

    @auto_field
    def numberOfPartialBathrooms(self) -> Optional[int]:
        return self.real_estate.numberOfPartialBathrooms

    @auto_field
    def numberOfRooms(self) -> Optional[int]:
        return self.real_estate.numberOfRooms

    @auto_field
    def price(self) -> Optional[str]:
        return self.real_estate.price

    @auto_field
    def propertyType(self) -> Optional[str]:
        return self.real_estate.propertyType

    @auto_field
    def realEstateId(self) -> Optional[str]:
        return self.real_estate.realEstateId

    @auto_field
    def rentalPeriod(self) -> Optional[str]:
        return self.real_estate.rentalPeriod

    @auto_field
    def tradeType(self) -> Optional[str]:
        return self.real_estate.tradeType

    @auto_field
    def url(self) -> Optional[str]:
        return self.real_estate.url

    @auto_field
    def virtualTourUrl(self) -> Optional[str]:
        return self.real_estate.virtualTourUrl

    @auto_field
    def yearBuilt(self) -> Optional[int]:
        return self.real_estate.yearBuilt



from typing import List, Optional

import attrs

from zyte_common_items.base import Item
from zyte_common_items.components import (
    AdditionalProperty,
    Address,
    AggregateRating,
    Amenity,
    Image,
    Metadata,
    NamedLink,
    OpeningHoursItem,
    ParentPlace,
    StarRating,
)
from zyte_common_items.converters import to_metadata_optional, url_to_str_optional


@attrs.define(kw_only=True)
class BusinessPlaceMetadata(Metadata):
    """Metadata class for :data:`zyte_common_items.BusinessPlace.metadata`."""


@attrs.define(slots=True, kw_only=True)
class BusinessPlace(Item):
    """Business place, with properties typically seen on maps or business
    listings.

    :attr:`url` is the only required attribute.
    """

    #: Unique identifier of the place on the website.
    placeId: Optional[str] = None

    #: The main URL that the place data was extracted from.
    #:
    #: The URL of the final response, after any redirects.
    #:
    #: In case there is no product data on the page or the page was not reached, the returned "empty"
    #: item would still contain url field and metadata field with dateDownloaded.
    url: Optional[str] = attrs.field(
        default=None, converter=url_to_str_optional, kw_only=True
    )

    #: The name of the place.
    name: Optional[str] = None

    #: List of actions that can be performed directly from the URLs on the place page, including URLs.
    actions: Optional[List[NamedLink]] = None

    #: List of name-value pais of any unmapped additional properties specific to the place.
    additionalProperties: Optional[List[AdditionalProperty]] = None

    #: The address details of the place.
    address: Optional[Address] = None

    #: The details of the reservation action,
    #: e.g. table reservation in case of restaurants
    #: or room reservation in case of hotels.
    reservationAction: Optional[NamedLink] = None

    #: List of categories the place belongs to.
    categories: Optional[List[str]] = None

    #: The description of the place.
    #:
    #: Stripped of white spaces.
    description: Optional[str] = None

    #: List of frequently mentioned features of this place.
    features: Optional[List[str]] = None

    #: URL to a map of the place.
    map: Optional[str] = attrs.field(
        default=None, converter=url_to_str_optional, kw_only=True
    )

    #: A list of URL values of all images of the place.
    images: Optional[List[Image]] = None

    #: List of amenities of the place.
    amenityFeatures: Optional[List[Amenity]] = None

    #: The overall rating, based on a collection of reviews or ratings.
    aggregateRating: Optional[AggregateRating] = None

    #: Official star rating of the place.
    starRating: Optional[StarRating] = None

    #: If the place is located inside another place, these are the details of the parent place.
    containedInPlace: Optional[ParentPlace] = None

    #: Ordered specification of opening hours, including data for opening and closing time for each day of the week.
    openingHours: Optional[List[OpeningHoursItem]] = None

    #: List of partner review sites.
    reviewSites: Optional[List[NamedLink]] = None

    #: The phone number associated with the place, as it appears on the page.
    telephone: Optional[str] = None

    #: How is the price range of the place viewed by its customers (from z to zzzz).
    priceRange: Optional[str] = None

    #: Which timezone is the place situated in.
    #:
    #: Standard: Name compliant with IANA tz database (tzdata).
    timezone: Optional[str] = None

    #: If the information is verified by the owner of this place.
    isVerified: Optional[bool] = None

    #: The URL pointing to the official website of the place.
    website: Optional[str] = attrs.field(
        default=None, converter=url_to_str_optional, kw_only=True
    )

    #:  List of the tags associated with the place.
    tags: Optional[List[str]] = None

    #: Data extraction process metadata.
    metadata: Optional[BusinessPlaceMetadata] = attrs.field(
        default=None, converter=to_metadata_optional(BusinessPlaceMetadata), kw_only=True  # type: ignore[misc]
    )



from typing import Optional

import attrs

from zyte_common_items.base import Item


@attrs.define(kw_only=True)
class Address(Item):
    """Address item."""

    #: The raw address information, as it appears on the website.
    addressRaw: Optional[str] = None

    #: The street address of the place.
    streetAddress: Optional[str] = None

    #: The city the place is located in.
    addressCity: Optional[str] = None

    #: The locality to which the place belongs.
    addressLocality: Optional[str] = None

    #: The region of the place.
    addressRegion: Optional[str] = None

    #: The country the place is located in.
    #:
    #: The country name or the `ISO 3166-1 alpha-2 country code
    #: <https://en.wikipedia.org/wiki/ISO_3166-1>`__.
    addressCountry: Optional[str] = None

    #: The postal code of the address.
    postalCode: Optional[str] = None

    #: The auxiliary part of the postal code.
    #:
    #: It may include a state abbreviation or town name, depending on local standards.
    postalCodeAux: Optional[str] = None

    #: Geographical latitude of the place.
    latitude: Optional[float] = None

    #: Geographical longitude of the place.
    longitude: Optional[float] = None



import base64
from typing import List, Optional, Type, TypeVar

import attrs

from zyte_common_items.base import Item, ProbabilityMixin
from zyte_common_items.components.metadata import ProbabilityMetadata
from zyte_common_items.util import convert_to_class, url_to_str


@attrs.define(kw_only=True)
class Header(Item):
    """An HTTP header"""

    #: Name of the header
    name: str

    #: Value of the header
    value: str


#: :class:`~typing.TypeVar` for :class:`Request`.
RequestT = TypeVar("RequestT", bound="Request")


@attrs.define(slots=False)
class Request(Item):
    """Describe a web request to load a page"""

    #: HTTP URL
    url: str = attrs.field(converter=url_to_str)

    #: HTTP method
    method: str = "GET"

    #: HTTP request body, Base64-encoded
    body: Optional[str] = None

    #: HTTP headers
    headers: Optional[List[Header]] = None

    #: Name of the page being requested.
    name: Optional[str] = None

    _body_bytes = None

    @property
    def body_bytes(self) -> Optional[bytes]:
        """Request.body as bytes"""
        # todo: allow to set body bytes in __init__, to avoid encoding/decoding.
        if self._body_bytes is None:
            if self.body is not None:
                self._body_bytes = base64.b64decode(self.body)
        return self._body_bytes

    def to_scrapy(self, callback, **kwargs):
        """
        Convert a request to scrapy.Request.
        All kwargs are passed to scrapy.Request as-is.
        """
        import scrapy

        header_list = [(header.name, header.value) for header in self.headers or []]

        return scrapy.Request(
            url=self.url,
            callback=callback,
            method=self.method or "GET",
            headers=header_list,
            body=self.body_bytes,
            **kwargs,
        )

    def cast(self, cls: Type[RequestT]) -> RequestT:
        """Convert *value*, an instance of :class:`~.Request` or a subclass, into
        *cls*, a different class that is also either :class:`~.Request` or a
        subclass."""
        new_value = convert_to_class(self, cls)
        if type(self) is Request and cls is ProbabilityRequest:
            new_value.metadata = ProbabilityMetadata(probability=1.0)  # type: ignore[attr-defined]
        return new_value


@attrs.define(kw_only=True)
class ProbabilityRequest(Request, ProbabilityMixin):
    """A :class:`Request` that includes a probability value."""

    #: Data extraction process metadata.
    metadata: Optional[ProbabilityMetadata] = None



import attrs

from zyte_common_items.base import Item


@attrs.define
class Gtin(Item):
    """GTIN_ type-value pair.

    See :class:`Product.gtin <zyte_common_items.Product.gtin>`.

    .. _GTIN: https://en.wikipedia.org/wiki/Global_Trade_Item_Number
    """

    #: Identifier of the GTIN format of ``value``.
    #:
    #: One of: ``"gtin13"``, ``"gtin8"``, ``"gtin14"``, ``"isbn10"``,
    #: ``"isbn13"``, ``"ismn"``, ``"issn"``, ``"upc"``.
    type: str

    #: Value.
    #:
    #: It should only contain digits.
    value: str



from copy import copy
from typing import Any, Dict, Type

import attrs
import pytest
from itemadapter import ItemAdapter
from web_poet import RequestUrl, Returns, field

from zyte_common_items import (
    Article,
    ArticleList,
    ArticleNavigation,
    AutoArticleListPage,
    AutoArticleNavigationPage,
    AutoArticlePage,
    AutoBusinessPlacePage,
    AutoJobPostingPage,
    AutoProductListPage,
    AutoProductNavigationPage,
    AutoProductPage,
    AutoRealEstatePage,
    BusinessPlace,
    JobPosting,
    Product,
    ProductList,
    ProductNavigation,
    RealEstate,
)

from .test_items import (
    _ARTICLE_ALL_KWARGS,
    _ARTICLE_LIST_ALL_KWARGS,
    _ARTICLE_NAVIGATION_ALL_KWARGS,
    _BUSINESS_PLACE_ALL_KWARGS,
    _JOB_POSTING_ALL_KWARGS,
    _PRODUCT_ALL_KWARGS,
    _PRODUCT_LIST_ALL_KWARGS,
    _PRODUCT_NAVIGATION_ALL_KWARGS,
    _REAL_ESTATE_ALL_KWARGS,
)

PARAMS = (
    "item_cls,item_kwargs,cls,param",
    (
        (Article, _ARTICLE_ALL_KWARGS, AutoArticlePage, "article"),
        (ArticleList, _ARTICLE_LIST_ALL_KWARGS, AutoArticleListPage, "article_list"),
        (
            ArticleNavigation,
            _ARTICLE_NAVIGATION_ALL_KWARGS,
            AutoArticleNavigationPage,
            "article_navigation",
        ),
        (
            BusinessPlace,
            _BUSINESS_PLACE_ALL_KWARGS,
            AutoBusinessPlacePage,
            "business_place",
        ),
        (
            JobPosting,
            _JOB_POSTING_ALL_KWARGS,
            AutoJobPostingPage,
            "job_posting",
        ),
        (Product, _PRODUCT_ALL_KWARGS, AutoProductPage, "product"),
        (ProductList, _PRODUCT_LIST_ALL_KWARGS, AutoProductListPage, "product_list"),
        (
            ProductNavigation,
            _PRODUCT_NAVIGATION_ALL_KWARGS,
            AutoProductNavigationPage,
            "product_navigation",
        ),
        (
            RealEstate,
            _REAL_ESTATE_ALL_KWARGS,
            AutoRealEstatePage,
            "real_estate",
        ),
    ),
)


async def assert_expected_item(page, item):
    assert await page.to_item() == item
    # Ensure that all fields are sync.
    for k, v in ItemAdapter(item).items():
        assert getattr(page, k) == v


@pytest.mark.parametrize(*PARAMS)
@pytest.mark.asyncio
async def test_unmodified(
    item_cls: Type, item_kwargs: Dict[str, Any], cls: Type, param: str
) -> None:
    item = item_cls(**item_kwargs)
    kwargs = {
        param: item,
        "request_url": RequestUrl("https://example.com"),
    }
    page = cls(**kwargs)
    await assert_expected_item(page, item)


@pytest.mark.parametrize(*PARAMS)
@pytest.mark.asyncio
async def test_modified(
    item_cls: Type, item_kwargs: Dict[str, Any], cls: Type, param: str
) -> None:
    modified_url = "https://custom.example"

    class CustomPage(cls):
        @field
        def url(self):
            return modified_url

    item = item_cls(**item_kwargs)
    kwargs = {
        param: item,
        "request_url": RequestUrl("https://example.com"),
    }
    page = CustomPage(**kwargs)
    expected_item = copy(item)
    expected_item.url = modified_url
    await assert_expected_item(page, expected_item)


@pytest.mark.parametrize(*PARAMS)
@pytest.mark.asyncio
async def test_extended(
    item_cls: Type, item_kwargs: Dict[str, Any], cls: Type, param: str
) -> None:
    @attrs.define
    class ExtendedItem(item_cls):
        foo: str

    class ExtendedPage(cls, Returns[ExtendedItem]):
        @field
        def foo(self):
            return "bar"

    kwargs = {
        param: item_cls(**item_kwargs),
        "request_url": RequestUrl("https://example.com"),
    }
    page = ExtendedPage(**kwargs)
    expected_item = ExtendedItem(**item_kwargs, foo="bar")
    await assert_expected_item(page, expected_item)



from typing import List, Optional

import attrs

from zyte_common_items.base import Item
from zyte_common_items.components import (
    BaseSalary,
    HiringOrganization,
    JobLocation,
    Metadata,
)
from zyte_common_items.converters import to_metadata_optional, url_to_str


@attrs.define(kw_only=True)
class JobPostingMetadata(Metadata):
    """Metadata class for :data:`zyte_common_items.JobPosting.metadata`."""


@attrs.define(kw_only=True)
class JobPosting(Item):
    """A job posting, typically seen on job posting websites or websites of
    companies that are hiring.

    :attr:`url` is the only required attribute.
    """

    #: The url of the final response, after any redirects.
    url: str = attrs.field(converter=url_to_str)

    #: The identifier of the job posting.
    jobPostingId: Optional[str] = None

    #: Publication date of the job posting.
    #:
    #: Format: ISO 8601 format: "YYYY-MM-DDThh:mm:ssZ"
    #:
    #: With timezone, if available.
    datePublished: Optional[str] = None

    #: Same date as datePublished, but before parsing/normalization, i.e. as it appears on the website.
    datePublishedRaw: Optional[str] = None

    #: The date when the job posting was most recently modified.
    #:
    #: Format: ISO 8601 format: "YYYY-MM-DDThh:mm:ssZ"
    #:
    #: With timezone, if available.
    dateModified: Optional[str] = None

    #: Same date as dateModified, but before parsing/normalization, i.e. as it appears on the website.
    dateModifiedRaw: Optional[str] = None

    #: The date after which the job posting is not valid, e.g. the end of an offer.
    #:
    #: Format: ISO 8601 format: "YYYY-MM-DDThh:mm:ssZ"
    #:
    #: With timezone, if available.
    validThrough: Optional[str] = None

    #: Same date as validThrough, but before parsing/normalization, i.e. as it appears on the website.
    validThroughRaw: Optional[str] = None

    #: The title of the job posting.
    jobTitle: Optional[str] = None

    #: The headline of the job posting.
    headline: Optional[str] = None

    #: A (typically single) geographic location associated with the job position.
    jobLocation: Optional[JobLocation] = None

    #: A description of the job posting including sub-headings, with newline separators.
    #:
    #: Format:
    #:
    #: - trimmed (no whitespace at the beginning or the end of the description string),
    #:
    #: - line breaks included,
    #:
    #: - no length limit,
    #:
    #: - no normalization of Unicode characters.
    description: Optional[str] = None

    #: Simplified HTML of the description, including sub-headings, image captions and embedded content.
    descriptionHtml: Optional[str] = None

    #: Type of employment (e.g. full-time, part-time, contract, temporary, seasonal, internship).
    employmentType: Optional[str] = None

    #: The base salary of the job or of an employee in the proposed role.
    baseSalary: Optional[BaseSalary] = None

    #: Candidate requirements for the job.
    requirements: Optional[List[str]] = None

    #: Information about the organization offering the job position.
    hiringOrganization: Optional[HiringOrganization] = None

    #: Job start date
    #:
    #: Format: ISO 8601 format: "YYYY-MM-DDThh:mm:ssZ"
    #:
    #: With timezone, if available.
    jobStartDate: Optional[str] = None

    #: Same date as jobStartDate, but before parsing/normalization, i.e. as it appears on the website.
    jobStartDateRaw: Optional[str] = None

    #: Specifies the remote status of the position.
    remoteStatus: Optional[str] = None

    #: Contains metadata about the data extraction process.
    metadata: Optional[JobPostingMetadata] = attrs.field(
        default=None, converter=to_metadata_optional(JobPostingMetadata), kw_only=True  # type: ignore[misc]
    )



import pytest

from zyte_common_items import Article, ArticleList, JobPosting, Product, ProductList
from zyte_common_items.ae import (
    AEGTIN,
    AEAdditionalProperty,
    AEArticle,
    AEArticleFromList,
    AEArticleList,
    AEBreadcrumb,
    AEJobPosting,
    AELocation,
    AEOffer,
    AEOrganization,
    AEPaginationLink,
    AEProduct,
    AEProductFromList,
    AEProductList,
    AERating,
    AESalary,
)
from zyte_common_items.pipelines import AEPipeline

from .test_items import (
    _ARTICLE_ALL_KWARGS,
    _ARTICLE_LIST_ALL_KWARGS,
    _ARTICLE_LIST_MIN_KWARGS,
    _ARTICLE_MIN_KWARGS,
    _JOB_POSTING_ALL_KWARGS,
    _JOB_POSTING_MIN_KWARGS,
    _PRODUCT_ALL_KWARGS,
    _PRODUCT_LIST_ALL_KWARGS,
    _PRODUCT_LIST_MIN_KWARGS,
    _PRODUCT_MIN_KWARGS,
)


@pytest.mark.parametrize(
    ["item", "expected"],
    (
        (
            Article(**_ARTICLE_MIN_KWARGS),
            AEArticle(
                probability=1.0,
                url="https://www.zyte.com/blog/product-data-extraction-automatic",
            ),
        ),
        (
            Article(**_ARTICLE_ALL_KWARGS),
            AEArticle(
                headline="Guarantee the best results for product data extraction",
                datePublished="2022-08-26T20:56:16Z",
                datePublishedRaw="August 26, 2022",
                dateModified="2022-08-27T21:48:55Z",
                dateModifiedRaw="August 25, 2022",
                inLanguage="en",
                breadcrumbs=[
                    AEBreadcrumb(name="Homepage"),
                    AEBreadcrumb(name="Blog"),
                    AEBreadcrumb(
                        name="Guarantee the best results for product data extraction",
                    ),
                ],
                mainImage="https://www.zyte.com/wp-content/uploads/2022/08/product-data-extraction-zyte.png",
                images=[
                    "https://www.zyte.com/wp-content/uploads/2022/08/product-data-extraction-zyte.png",
                    "https://www.zyte.com/wp-content/uploads/2021/06/Webinar_Product-Data-Quality-01-1024x536.png",
                    "https://www.zyte.com/wp-content/uploads/2022/08/product-data-extraction-automatic.png",
                ],
                description="Product Data Extraction helps understand how consumers use a specific product, foresee improvements and adjustments that ultimately increase sales and demand.",
                articleBody="When businesses operate in a competitive environment it is imperative to know what their competitors are charging in real-time and this can be hard to keep track of. For any data driven organization, implementing a solution that automatically extracts product data from websites in real-time and at scale, is indispensable to stay ahead of the competition. Setting an automatic process for product data extraction can be a powerful tool for data driven businesses of all sizes. You can extract specific products information including offers, price, currency, and availability. Provided that the data extraction process is able to identify the key attributes of a product, it can then use this information to create reports with insights into the behavior of a particular product. For this to work, it is key to have the right procedures in place that guarantee the best results for your automation product extraction projects. By doing so, your organization can better understand how consumers are using a specific product, foresee any necessary adjustments to improve the user experience and as a result, increase sales and demand. In this article, we'll explain not just everything you need to know to get started, but also show you how to guarantee the best results when working with product data extraction.\n\nWhat is automatic product extraction?\n\n...",
                articleBodyHtml='<article>\n<p>When businesses operate in a competitive environment it is imperative to know what their competitors are charging in real-time and this can be hard to keep track of. For any data driven organization, implementing a solution that automatically extracts product data from websites in real-time and at scale, is indispensable to stay ahead of the competition.</p> <p>Setting an automatic process for <a href="https://docs.zyte.com/automatic-extraction/product.html">product data extraction</a> can be a powerful tool for data driven businesses of all sizes. You can extract specific products information including offers, price, currency, and availability. Provided that the data extraction process is able to <a href="/blog/automatic-extraction-data-extractor-review/">identify the key attributes of a product</a>, it can then use this information to create reports with insights into the behavior of a particular product.</p>\n<p>For this to work, it is key to have the right procedures in place that guarantee the best results for your automation product extraction projects. By doing so, your organization can better understand how consumers are using a specific product, foresee any necessary adjustments to improve the user experience and as a result, increase sales and demand.</p>\n<p>In this article, we\'ll explain not just everything you need to know to get started, but also show you how to guarantee the best results when working with product data extraction.</p>\n<h2><strong>What is automatic product extraction?</strong></h2>\n...\n</article>',
                videoUrls=[
                    "https://player.vimeo.com/video/90592864?title=0&byline=0&portrait=0&color=8dc7dc"
                ],
                audioUrls=["https://example.com/audiofile.mp3"],
                probability=1.0,
                canonicalUrl="https://www.zyte.com/blog/product-data-extraction-automatic/",
                url="https://www.zyte.com/blog/product-data-extraction-automatic",
            ),
        ),
        (
            ArticleList(**_ARTICLE_LIST_MIN_KWARGS),
            AEArticleList(
                url="https://www.zyte.com/blog/extract-summit-blog/",
            ),
        ),
        (
            ArticleList(**_ARTICLE_LIST_ALL_KWARGS),
            AEArticleList(
                url="https://www.zyte.com/blog/extract-summit-blog/",
                articles=[
                    AEArticleFromList(
                        url="https://www.zyte.com/blog/reflecting-on-the-2022-web-data-extraction-summit-zyte/",
                        probability=1.0,
                    )
                ],
            ),
        ),
        (
            JobPosting(**_JOB_POSTING_MIN_KWARGS),
            AEJobPosting(
                probability=1.0,
                url="https://example.com/viewjob/12345",
            ),
        ),
        (
            JobPosting(**_JOB_POSTING_ALL_KWARGS),
            AEJobPosting(
                title="Software Engineer",
                datePosted="2019-06-19T00:00:00-05:00",
                validThrough="2019-07-19T00:00:00-05:00",
                description="We are looking for a Software Engineer to join our team.- 35 days holiday- 15% bonus- flexible working arrangements",
                descriptionHtml="<p>We are looking for a Software Engineer to join our team.</p><ul><li>35 days holiday</li><li>15% bonus</li><li>flexible working arrangements</li></ul>",
                employmentType="Full-time",
                hiringOrganization=AEOrganization(raw="ACME Corp."),
                baseSalary=AESalary(
                    raw="$53,000-$55,000 a year", value=55000.0, currency="$"
                ),
                jobLocation=AELocation(raw="New York, NY"),
                probability=0.95,
                url="https://example.com/viewjob/12345",
            ),
        ),
        (
            Product(**_PRODUCT_MIN_KWARGS),
            AEProduct(
                probability=1.0,
                url="https://example.com/?product=product22",
            ),
        ),
        (
            Product(**_PRODUCT_ALL_KWARGS),
            AEProduct(
                url="https://example.com/?product=product22",
                canonicalUrl="https://example.com/product22",
                probability=1.0,
                name="White two-door refrigerator",
                offers=[
                    AEOffer(
                        price="9999.99",
                        currency="$",
                        availability="InStock",
                        regularPrice="11999.99",
                    ),
                ],
                sku="A123DK9823",
                gtin=[
                    AEGTIN(
                        type="foo",
                        value="bar",
                    ),
                ],
                mpn="HSC0424PP",
                brand="Ka-pow",
                breadcrumbs=[
                    AEBreadcrumb(
                        name="Level 1",
                    ),
                    AEBreadcrumb(
                        name="Level 2",
                    ),
                ],
                mainImage="http://example.com/image1.png",
                images=[
                    "http://example.com/image1.png",
                ],
                description=(
                    "Full freshness all over the fridge\n"
                    "5 Conversion Modes on demand\n"
                    "Super Cooling Plus™"
                ),
                descriptionHtml=(
                    "<article>"
                    "<p>Full freshness all over the fridge</p>"
                    "<p>5 Conversion Modes on demand</p>"
                    "<p>Super Cooling Plus&trade;</p>"
                    "</article>"
                ),
                additionalProperty=[
                    AEAdditionalProperty(
                        name="foo",
                        value="bar",
                    ),
                ],
                aggregateRating=AERating(
                    ratingValue=2.5,
                    bestRating=5.0,
                    reviewCount=123,
                ),
                color="white",
                size="XL",
                style="polka dots",
                hasVariants=[
                    AEProduct(
                        url="https://example.com/?product=product22",
                        probability=1.0,
                    )
                ],
            ),
        ),
        (
            ProductList(**_PRODUCT_LIST_MIN_KWARGS),
            AEProductList(
                url="https://example.com/swiss-watches?sort=new-first",
            ),
        ),
        (
            ProductList(**_PRODUCT_LIST_ALL_KWARGS),
            AEProductList(
                url="https://example.com/swiss-watches?sort=new-first",
                products=[AEProductFromList(probability=1.0)],
                breadcrumbs=[
                    AEBreadcrumb(name="Level 1"),
                    AEBreadcrumb(name="Level 2"),
                ],
                paginationNext=AEPaginationLink(
                    url="https://example.com/swiss-watches?sort=new-first&page=2",
                    text="foo",
                ),
            ),
        ),
    ),
)
def test_main(item, expected):
    with pytest.deprecated_call():
        pipeline = AEPipeline()
    actual = pipeline.process_item(item, spider=None)
    assert expected == actual



from datetime import datetime
from typing import Dict, List, Optional, Type, TypeVar

import attrs

from zyte_common_items import Item
from zyte_common_items._dateutils import parse_iso_datetime
from zyte_common_items.util import convert_to_class

#: :class:`~typing.TypeVar` for :class:`BaseMetadata`.
MetadataT = TypeVar("MetadataT", bound="BaseMetadata")


class BaseMetadata(Item):
    """Base metadata class"""

    def cast(self, cls: Type[MetadataT]) -> MetadataT:
        """Convert *value*, a metadata instance, into a different metadata *cls*."""
        return convert_to_class(self, cls)


@attrs.define(kw_only=True)
class ProbabilityMetadata(BaseMetadata):
    """Data extraction process metadata."""

    #: The probability (0 for 0%, 1 for 100%) that the resource features the
    #: expected data type.
    #:
    #: For example, if the extraction of a product from a given URL is
    #: requested, and that URL points to the webpage of a product with complete
    #: certainty, the value should be `1`. If with complete certainty the
    #: webpage features a job listing instead of a product, the value should be
    #: `0`. When there is no complete certainty, the value could be anything in
    #: between (e.g. `0.96`).
    probability: Optional[float] = 1.0


@attrs.define(kw_only=True)
class ListMetadata(BaseMetadata):
    """Minimal metadata for list item classes, such as ProductList or ArticleList.

    See
    :class:`ArticleList.metadata <zyte_common_items.ArticleList.metadata>`.
    """

    #: Date and time when the product data was downloaded, in UTC timezone and
    #: the following format: ``YYYY-MM-DDThh:mm:ssZ``.
    dateDownloaded: Optional[str] = None

    #: Contains paths to fields with the description of issues found with their values.
    validationMessages: Optional[Dict[str, List[str]]] = None

    def get_date_downloaded_parsed(self) -> Optional[datetime]:
        """Return dateDownloaded as a TZ-aware datetime object"""
        if self.dateDownloaded is None:
            return None
        return parse_iso_datetime(self.dateDownloaded)


@attrs.define(kw_only=True)
class DetailsMetadata(ListMetadata):
    """Minimal metadata for details item classes, such as Product or Article."""

    #: The probability (0 for 0%, 1 for 100%) that the resource features the
    #: expected data type.
    #:
    #: For example, if the extraction of a product from a given URL is
    #: requested, and that URL points to the webpage of a product with complete
    #: certainty, the value should be `1`. If with complete certainty the
    #: webpage features a job listing instead of a product, the value should be
    #: `0`. When there is no complete certainty, the value could be anything in
    #: between (e.g. `0.96`).
    probability: Optional[float] = 1.0


@attrs.define(kw_only=True)
class Metadata(DetailsMetadata):
    """Generic metadata class.

    It defines all attributes of metadata classes for specific item types, so
    that it can be used during extraction instead of a more specific class, and
    later converted to the corresponding, more specific metadata class.
    """

    #: The search text used to find the item.
    searchText: Optional[str] = None



import json
from base64 import b64encode
from copy import copy
from importlib.metadata import version

import attrs
import pytest
from itemadapter import ItemAdapter
from packaging.version import Version
from w3lib.url import add_or_replace_parameters
from web_poet import RequestUrl, field

from zyte_common_items import Header, Request, SearchRequestTemplatePage


@pytest.mark.skipif(
    Version(version("jinja2")) < Version("2.9"),
    reason="The tojson filter was added in Jinja 2.9",
)
@pytest.mark.asyncio
async def test_all():
    class HolisticSearchRequestTemplatePage(SearchRequestTemplatePage):
        @field
        def url(self):
            return """
                {%-
                    if keyword|length > 1
                    and keyword[0]|lower == 'p'
                    and keyword[1:]|int(-1) != -1
                -%}
                    https://example.com/p/{{ keyword|upper }}
                {%- else -%}
                    https://example.com/search
                {%- endif -%}
            """

        @field
        def method(self):
            return """
                {%-
                    if keyword|length > 1
                    and keyword[0]|lower == 'p'
                    and keyword[1:]|int(-1) != -1
                -%}
                    GET
                {%- else -%}
                    POST
                {%- endif -%}
            """

        @field
        def body(self):
            return """
                {%-
                    if keyword|length > 1
                    and keyword[0]|lower == 'p'
                    and keyword[1:]|int(-1) != -1
                -%}
                {%- else -%}
                    {"query": {{ keyword|tojson }}}
                {%- endif -%}
            """

        @field
        def headers(self):
            return [
                Header(
                    name=(
                        """
                            {%-
                                if keyword|length > 1
                                and keyword[0]|lower == 'p'
                                and keyword[1:]|int(-1) != -1
                            -%}
                            {%- else -%}
                                Query
                            {%- endif -%}
                        """
                    ),
                    value="{{ keyword }}",
                ),
            ]

    search_request_template = await HolisticSearchRequestTemplatePage().to_item()

    search_request = search_request_template.request(keyword="p250")
    expected_request = Request("https://example.com/p/P250")
    assert search_request == expected_request

    search_request = search_request_template.request(keyword="foo bar")
    expected_request = Request(
        "https://example.com/search",
        method="POST",
        body=b64encode(json.dumps({"query": "foo bar"}).encode()).decode(),
        headers=[
            Header(name="Query", value="foo bar"),
        ],
    )
    assert search_request == expected_request


class VerbatimSearchRequestTemplatePage(SearchRequestTemplatePage):
    @field
    def url(self):
        return "https://example.com/?search={{ keyword }}"


class QuoteSearchRequestTemplatePage(SearchRequestTemplatePage):
    @field
    def url(self):
        return "https://example.com/?search={{ keyword|urlencode }}"


class QuotePlusSearchRequestTemplatePage(SearchRequestTemplatePage):
    @field
    def url(self):
        return "https://example.com/?search={{ keyword|quote_plus }}"


class ReplaceSearchRequestTemplatePage(SearchRequestTemplatePage):
    @field
    def url(self):
        return "https://example.com/search/{{ keyword|replace(' ', '/') }}"


@attrs.define
class UrlBasedSearchRequestTemplatePage(SearchRequestTemplatePage):
    request_url: RequestUrl

    @field
    def url(self):
        return f"{self.request_url}?search={{{{ keyword|urlencode }}}}"


# Defines way more of what the corresponding test scenario uses to give an idea
# of the future possibilities of this approach.
def edit_request_url(expression, page):
    if isinstance(expression, str):
        return expression
    if not isinstance(expression, dict):
        raise ValueError(
            f"The edit_request_url processor expected a dict, got " f"{expression!r}"
        )
    if "url" in expression:
        url = expression["url"]
    else:
        adapter = ItemAdapter(page)
        for v in adapter.values():
            if isinstance(v, RequestUrl):
                url = str(v)
                break
        else:
            raise ValueError(
                f"Cannot determine the base URL. Expression {expression!r} "
                f"for the edit_request_url processor has no 'url' key, and "
                f"the parent page object class does not define any RequestUrl "
                f"dependency."
            )
    url_safe_keyword_placeholder = "7dnKEYWORD2Ua"
    if "add_query_params" in expression:
        params = copy(expression["add_query_params"])
        for k in list(params):
            v = params.pop(k)
            k = k.format(keyword=url_safe_keyword_placeholder)
            v = v.format(keyword=url_safe_keyword_placeholder)
            params[k] = v
        url = add_or_replace_parameters(url, params)
    url = url.replace(url_safe_keyword_placeholder, "{{ keyword|urlencode }}")
    return url


@attrs.define
class DSLSearchRequestTemplatePage(SearchRequestTemplatePage):
    request_url: RequestUrl

    class Processors:
        url = [edit_request_url]

    @field
    def url(self):
        return {"add_query_params": {"search": "{keyword}"}}


@pytest.mark.parametrize(
    ("page", "inputs", "keyword", "url"),
    (
        (
            VerbatimSearchRequestTemplatePage,
            {},
            "foo bar",
            "https://example.com/?search=foo bar",
        ),
        (
            QuoteSearchRequestTemplatePage,
            {},
            "foo bar",
            "https://example.com/?search=foo%20bar",
        ),
        (
            QuotePlusSearchRequestTemplatePage,
            {},
            "foo bar",
            "https://example.com/?search=foo+bar",
        ),
        (
            ReplaceSearchRequestTemplatePage,
            {},
            "foo bar",
            "https://example.com/search/foo/bar",
        ),
        (
            UrlBasedSearchRequestTemplatePage,
            {"request_url": RequestUrl("https://foo.example/")},
            "foo bar",
            "https://foo.example/?search=foo%20bar",
        ),
        (
            DSLSearchRequestTemplatePage,
            {"request_url": RequestUrl("https://example.com/")},
            "foo bar",
            "https://example.com/?search=foo%20bar",
        ),
    ),
)
@pytest.mark.asyncio
async def test_url(page, inputs, keyword, url):
    search_request_template = await page(**inputs).to_item()
    search_request = search_request_template.request(keyword=keyword)
    assert search_request.url == url


@pytest.mark.asyncio
async def test_body_space():
    class BodySpaceSearchRequestTemplatePage(SearchRequestTemplatePage):
        @field
        def url(self):
            return "https://example.com"

        @field
        def body(self):
            return " "

    search_request_template = await BodySpaceSearchRequestTemplatePage().to_item()
    search_request = search_request_template.request(keyword="foo bar")
    assert search_request.body == b64encode(b" ").decode()


@pytest.mark.asyncio
async def test_header_empty_value():
    class BodySpaceSearchRequestTemplatePage(SearchRequestTemplatePage):
        @field
        def url(self):
            return "https://example.com"

        @field
        def headers(self):
            return [Header(name="Foo", value="")]

    search_request_template = await BodySpaceSearchRequestTemplatePage().to_item()
    search_request = search_request_template.request(keyword="foo bar")
    assert search_request.headers == [Header(name="Foo", value="")]



from __future__ import annotations

from typing import List, Optional

import attrs

from zyte_common_items.base import Item
from zyte_common_items.components import (
    AdditionalProperty,
    AggregateRating,
    Brand,
    Breadcrumb,
    DetailsMetadata,
    Gtin,
    Image,
)
from zyte_common_items.converters import (
    to_metadata_optional,
    url_to_str,
    url_to_str_optional,
)


@attrs.define(kw_only=True)
class ProductMetadata(DetailsMetadata):
    """Metadata class for :data:`zyte_common_items.Product.metadata`."""


@attrs.define(kw_only=True)
class Product(Item):
    """Product from an e-commerce website.

    :attr:`url` is the only required attribute.
    """

    #: List of name-value pais of data about a specific, otherwise unmapped
    #: feature.
    #:
    #: Additional properties usually appear in product pages in the form of a
    #: specification table or a free-form specification list.
    #:
    #: Additional properties that require 1 or more extra requests may not be
    #: extracted.
    #:
    #: See also ``features``.
    additionalProperties: Optional[List[AdditionalProperty]] = None

    #: Aggregate data about reviews and ratings.
    aggregateRating: Optional[AggregateRating] = None

    #: Availability status.
    #:
    #: The value is expected to be one of: ``"InStock"``, ``"OutOfStock"``.
    availability: Optional[str] = None

    #: Brand.
    brand: Optional[Brand] = None

    #: Webpage `breadcrumb trail`_.
    #:
    #: .. _Breadcrumb trail: https://en.wikipedia.org/wiki/Breadcrumb_navigation
    breadcrumbs: Optional[List[Breadcrumb]] = None

    #: Canonical form of the URL, as indicated by the website.
    #:
    #: See also ``url``.
    canonicalUrl: Optional[str] = attrs.field(
        default=None, converter=url_to_str_optional, kw_only=True
    )

    #: Color.
    #:
    #: It is extracted as displayed (e.g. ``"white"``).
    #:
    #: See also ``size``, ``style``.
    color: Optional[str] = None

    #: Price currency `ISO 4217`_ alphabetic code (e.g. ``"USD"``).
    #:
    #: See also ``currencyRaw``.
    #:
    #: .. _ISO 4217: https://en.wikipedia.org/wiki/ISO_4217
    currency: Optional[str] = None

    #: Price currency as it appears on the webpage (no post-processing), e.g.
    #: ``"$"``.
    #:
    #: See also ``currency``.
    currencyRaw: Optional[str] = None

    #: Plain-text description.
    #:
    #: If the description is split across different parts of the source
    #: webpage, only the main part, containing the most useful pieces of
    #: information, should be extracted into this attribute.
    #:
    #: It may contain data found in other attributes (``features``,
    #: ``additionalProperties``).
    #:
    #: Format-wise:
    #:
    #: -   Line breaks and non-ASCII characters are allowed.
    #:
    #: -   There is no length limit for this attribute, the content should not
    #:     be truncated.
    #:
    #: -   There should be no whitespace at the beginning or end.
    #:
    #: See also ``descriptionHtml``.
    description: Optional[str] = None

    #: HTML description.
    #:
    #: See ``description`` for extraction details.
    #:
    #: The format is not the raw HTML from the source webpage. See the `HTML
    #: normalization specification`_ for details.
    #:
    #: .. _HTML normalization specification: https://docs.zyte.com/automatic-extraction/article.html#format-of-articlebodyhtml-field
    descriptionHtml: Optional[str] = None

    #: List of features.
    #:
    #: They are usually listed as bullet points in product webpages.
    #:
    #: See also ``additionalProperties``.
    features: Optional[List[str]] = None

    #: List of standardized GTIN_ product identifiers associated with the
    #: product, which are unique for the product across different sellers.
    #:
    #: See also: ``mpn``, ``productId``, ``sku``.
    #:
    #: .. _GTIN: https://en.wikipedia.org/wiki/Global_Trade_Item_Number
    gtin: Optional[List[Gtin]] = None

    #: All product images.
    #:
    #: The main image (see ``mainImage``) should be first in the list.
    #:
    #: Images only displayed as part of the product description are excluded.
    images: Optional[List[Image]] = None

    #: Main product image.
    mainImage: Optional[Image] = None

    #: Data extraction process metadata.
    metadata: Optional[ProductMetadata] = attrs.field(
        default=None, converter=to_metadata_optional(ProductMetadata), kw_only=True  # type: ignore[misc]
    )

    #: `Manufacturer part number (MPN)`_.
    #:
    #: A product should have the same MPN across different e-commerce websites.
    #:
    #: See also: ``gtin``, ``productId``, ``sku``.
    #:
    #: .. _Manufacturer part number (MPN): https://en.wikipedia.org/wiki/Part_number
    mpn: Optional[str] = None

    #: Name as it appears on the webpage (no post-processing).
    name: Optional[str] = None

    #: Price at which the product is being offered.
    #:
    #: It is a string with the price amount, with a full stop as decimal
    #: separator, and no thousands separator or currency (see ``currency`` and
    #: ``currencyRaw``), e.g. ``"10500.99"``.
    #:
    #: If ``regularPrice`` is not ``None``, ``price`` should always be lower
    #: than ``regularPrice``.
    price: Optional[str] = None

    # Redefined to extend the documentation.
    #: Product identifier, unique within an e-commerce website.
    #:
    #: It may come in the form of an SKU or any other identifier, a hash, or
    #: even a URL.
    #:
    #: See also: ``gtin``, ``mpn``, ``sku``.
    productId: Optional[str] = None

    #: Price at which the product was being offered in the past, and which is
    #: presented as a reference next to the current price.
    #:
    #: It may be labeled as the original price, the list price, or the maximum
    #: retail price for which the product is sold.
    #:
    #: See ``price`` for format details.
    #:
    #: If ``regularPrice`` is not ``None``, it should always be higher than
    #: ``price``.
    regularPrice: Optional[str] = None

    #: Size or dimensions.
    #:
    #: Pertinent to products such as garments, shoes, accessories, etc.
    #:
    #: It is extracted as displayed (e.g. ``"XL"``).
    #:
    #: See also ``color``, ``style``.
    size: Optional[str] = None

    #: `Stock keeping unit (SKU)`_ identifier, i.e. a merchant-specific product
    #: identifier.
    #:
    #: See also: ``gtin``, ``mpn``, ``productId``.
    #:
    #: .. _Stock keeping unit (SKU): https://en.wikipedia.org/wiki/Stock_keeping_unit
    sku: Optional[str] = None

    #: Style.
    #:
    #: Pertinent to products such as garments, shoes, accessories, etc.
    #:
    #: It is extracted as displayed (e.g. ``"polka dots"``).
    #:
    #: See also ``color``, ``size``.
    style: Optional[str] = None

    #: Main URL from which the data has been extracted.
    #:
    #: See also ``canonicalUrl``.
    url: str = attrs.field(converter=url_to_str)

    #: List of variants.
    #:
    #: When slightly different versions of a product are displayed on the same
    #: product page, allowing you to choose a specific product version from a
    #: selection, each of those product versions are considered a product
    #: variant.
    #:
    #: Product variants usually differ in ``color`` or ``size``.
    #:
    #: The following items are *not* considered product variants:
    #:
    #: -   Different products within the same bundle of products.
    #:
    #: -   Product add-ons, e.g. premium upgrades of a base product.
    #:
    #: Only variant-specific data is extracted as product variant details. For
    #: example, if variant-specific versions of the product description do not
    #: exist in the source webpage, the description attributes of the product
    #: variant are *not* filled with the base product description.
    #:
    #: Extracted product variants may not include those that are not visible in
    #: the source webpage.
    #:
    #: Product variant details may not include those that require multiple
    #: additional requests (e.g. 1 or more requests per variant).
    variants: Optional[List[ProductVariant]] = None


@attrs.define(kw_only=True)
class ProductVariant(Item):
    """:class:`Product` variant.

    See :attr:`Product.variants`, :class:`ProductVariantExtractor`,
    :class:`ProductVariantSelectorExtractor`.
    """

    #: List of name-value pais of data about a specific, otherwise unmapped
    #: feature.
    #:
    #: Additional properties usually appear in product pages in the form of a
    #: specification table or a free-form specification list.
    #:
    #: Additional properties that require 1 or more extra requests may not be
    #: extracted.
    #:
    #: See also ``features``.
    additionalProperties: Optional[List[AdditionalProperty]] = None

    #: Availability status.
    #:
    #: The value is expected to be one of: ``"InStock"``, ``"OutOfStock"``.
    availability: Optional[str] = None

    #: Canonical form of the URL, as indicated by the website.
    #:
    #: See also ``url``.
    canonicalUrl: Optional[str] = attrs.field(
        default=None, converter=url_to_str_optional, kw_only=True
    )

    #: Color.
    #:
    #: It is extracted as displayed (e.g. ``"white"``).
    #:
    #: See also ``size``, ``style``.
    color: Optional[str] = None

    #: Price currency `ISO 4217`_ alphabetic code (e.g. ``"USD"``).
    #:
    #: See also ``currencyRaw``.
    #:
    #: .. _ISO 4217: https://en.wikipedia.org/wiki/ISO_4217
    currency: Optional[str] = None

    #: Price currency as it appears on the webpage (no post-processing), e.g.
    #: ``"$"``.
    #:
    #: See also ``currency``.
    currencyRaw: Optional[str] = None

    #: List of standardized GTIN_ product identifiers associated with the
    #: product, which are unique for the product across different sellers.
    #:
    #: See also: ``mpn``, ``productId``, ``sku``.
    #:
    #: .. _GTIN: https://en.wikipedia.org/wiki/Global_Trade_Item_Number
    gtin: Optional[List[Gtin]] = None

    #: All product images.
    #:
    #: The main image (see ``mainImage``) should be first in the list.
    #:
    #: Images only displayed as part of the product description are excluded.
    images: Optional[List[Image]] = None

    #: Main product image.
    mainImage: Optional[Image] = None

    #: `Manufacturer part number (MPN)`_.
    #:
    #: A product should have the same MPN across different e-commerce websites.
    #:
    #: See also: ``gtin``, ``productId``, ``sku``.
    #:
    #: .. _Manufacturer part number (MPN): https://en.wikipedia.org/wiki/Part_number
    mpn: Optional[str] = None

    #: Name as it appears on the webpage (no post-processing).
    name: Optional[str] = None

    #: Price at which the product is being offered.
    #:
    #: It is a string with the price amount, with a full stop as decimal
    #: separator, and no thousands separator or currency (see ``currency`` and
    #: ``currencyRaw``), e.g. ``"10500.99"``.
    #:
    #: If ``regularPrice`` is not ``None``, ``price`` should always be lower
    #: than ``regularPrice``.
    price: Optional[str] = None

    #: Product identifier, unique within an e-commerce website.
    #:
    #: It may come in the form of an SKU or any other identifier, a hash, or
    #: even a URL.
    #:
    #: See also: ``gtin``, ``mpn``, ``sku``.
    productId: Optional[str] = None

    #: Price at which the product was being offered in the past, and which is
    #: presented as a reference next to the current price.
    #:
    #: It may be labeled as the original price, the list price, or the maximum
    #: retail price for which the product is sold.
    #:
    #: See ``price`` for format details.
    #:
    #: If ``regularPrice`` is not ``None``, it should always be higher than
    #: ``price``.
    regularPrice: Optional[str] = None

    #: Size or dimensions.
    #:
    #: Pertinent to products such as garments, shoes, accessories, etc.
    #:
    #: It is extracted as displayed (e.g. ``"XL"``).
    #:
    #: See also ``color``, ``style``.
    size: Optional[str] = None

    #: `Stock keeping unit (SKU)`_ identifier, i.e. a merchant-specific product
    #: identifier.
    #:
    #: See also: ``gtin``, ``mpn``, ``productId``.
    #:
    #: .. _Stock keeping unit (SKU): https://en.wikipedia.org/wiki/Stock_keeping_unit
    sku: Optional[str] = None

    #: Style.
    #:
    #: Pertinent to products such as garments, shoes, accessories, etc.
    #:
    #: It is extracted as displayed (e.g. ``"polka dots"``).
    #:
    #: See also ``color``, ``size``.
    style: Optional[str] = None

    #: Main URL from which the product variant data could be extracted.
    #:
    #: See also ``canonicalUrl``.
    url: Optional[str] = attrs.field(
        default=None, converter=url_to_str_optional, kw_only=True
    )



from typing import Optional

import attrs

from zyte_common_items import Item
from zyte_common_items.util import url_to_str


@attrs.define(kw_only=True)
class Link(Item):
    """A link from a webpage to another webpage."""

    #: Displayed text.
    text: Optional[str] = None

    #: Target URL.
    url: Optional[str] = attrs.field(
        default=None, converter=attrs.converters.optional(url_to_str), kw_only=True
    )


@attrs.define(kw_only=True)
class NamedLink(Item):
    """A link from a webpage to another webpage."""

    #: The name of the link.
    name: Optional[str] = None

    #: Target URL.
    url: Optional[str] = attrs.field(
        default=None, converter=attrs.converters.optional(url_to_str), kw_only=True
    )


@attrs.define(kw_only=True)
class Url(Item):
    """A URL."""

    # : URL.
    url: str = attrs.field(converter=url_to_str)



from typing import List, Optional

import attrs
from web_poet import Returns

from zyte_common_items.components import (
    AdditionalProperty,
    Address,
    AggregateRating,
    Amenity,
    Image,
    NamedLink,
    OpeningHoursItem,
    ParentPlace,
    StarRating,
)
from zyte_common_items.fields import auto_field
from zyte_common_items.items import BusinessPlace, BusinessPlaceMetadata
from zyte_common_items.processors import (
    description_processor,
    images_processor,
    rating_processor,
)

from .base import BasePage, Page
from .mixins import HasMetadata


class BaseBusinessPlacePage(
    BasePage, Returns[BusinessPlace], HasMetadata[BusinessPlaceMetadata]
):
    """:class:`BasePage` subclass for :class:`BusinessPlace`."""

    class Processors(BasePage.Processors):
        aggregateRating = [rating_processor]
        description = [description_processor]
        images = [images_processor]


class BusinessPlacePage(
    Page, Returns[BusinessPlace], HasMetadata[BusinessPlaceMetadata]
):
    """:class:`Page` subclass for :class:`BusinessPlace`."""

    class Processors(Page.Processors):
        aggregateRating = [rating_processor]
        description = [description_processor]
        images = [images_processor]


@attrs.define
class AutoBusinessPlacePage(BaseBusinessPlacePage):
    business_place: BusinessPlace

    @auto_field
    def actions(self) -> Optional[List[NamedLink]]:
        return self.business_place.actions

    @auto_field
    def additionalProperties(self) -> Optional[List[AdditionalProperty]]:
        return self.business_place.additionalProperties

    @auto_field
    def address(self) -> Optional[Address]:
        return self.business_place.address

    @auto_field
    def aggregateRating(self) -> Optional[AggregateRating]:
        return self.business_place.aggregateRating

    @auto_field
    def amenityFeatures(self) -> Optional[List[Amenity]]:
        return self.business_place.amenityFeatures

    @auto_field
    def categories(self) -> Optional[List[str]]:
        return self.business_place.categories

    @auto_field
    def containedInPlace(self) -> Optional[ParentPlace]:
        return self.business_place.containedInPlace

    @auto_field
    def description(self) -> Optional[str]:
        return self.business_place.description

    @auto_field
    def features(self) -> Optional[List[str]]:
        return self.business_place.features

    @auto_field
    def images(self) -> Optional[List[Image]]:
        return self.business_place.images

    @auto_field
    def isVerified(self) -> Optional[bool]:
        return self.business_place.isVerified

    @auto_field
    def map(self) -> Optional[str]:
        return self.business_place.map

    @auto_field
    def metadata(self) -> Optional[BusinessPlaceMetadata]:
        return self.business_place.metadata

    @auto_field
    def name(self) -> Optional[str]:
        return self.business_place.name

    @auto_field
    def openingHours(self) -> Optional[List[OpeningHoursItem]]:
        return self.business_place.openingHours

    @auto_field
    def placeId(self) -> Optional[str]:
        return self.business_place.placeId

    @auto_field
    def priceRange(self) -> Optional[str]:
        return self.business_place.priceRange

    @auto_field
    def reservationAction(self) -> Optional[NamedLink]:
        return self.business_place.reservationAction

    @auto_field
    def reviewSites(self) -> Optional[List[NamedLink]]:
        return self.business_place.reviewSites

    @auto_field
    def starRating(self) -> Optional[StarRating]:
        return self.business_place.starRating

    @auto_field
    def tags(self) -> Optional[List[str]]:
        return self.business_place.tags

    @auto_field
    def telephone(self) -> Optional[str]:
        return self.business_place.telephone

    @auto_field
    def timezone(self) -> Optional[str]:
        return self.business_place.timezone

    @auto_field
    def url(self) -> Optional[str]:
        return self.business_place.url

    @auto_field
    def website(self) -> Optional[str]:
        return self.business_place.website



from copy import deepcopy
from unittest.mock import MagicMock, patch

import pytest

from zyte_common_items import Article, ArticleListPage, Product, ProductNavigation
from zyte_common_items.pipelines import DropLowProbabilityItemPipeline

scrapy = pytest.importorskip("scrapy")  # noqa


@pytest.mark.parametrize(
    "thresholds_settings, default_threshold, expected_thresholds, expected_default_thresholds",
    [
        ({}, 0.09, {}, 0.09),
        (
            {
                "zyte_common_items.Article": 0.2,
                "zyte_common_items.Product": 0.3,
            },
            None,
            {
                Article: 0.2,
                Product: 0.3,
            },
            None,
        ),  # thresholds for items per Item's path
        (
            {Article: 0.4, Product: 0.5},
            None,
            {
                Article: 0.4,
                Product: 0.5,
            },
            None,
        ),  # thresholds for items per Item's class
        (
            {Article: 0.4},
            0.09,
            {
                Article: 0.4,
            },
            0.09,
        ),  # default threshold is taken from DEFAULT_ITEM_PROBABILITY_THRESHOLD
        (
            {"default": 0.6, Article: 0.4},
            0.09,
            {
                Article: 0.4,
            },
            0.6,
        ),  # default threshold is taken from settings with the key 'default'
        (
            {"default": 0, Article: 0.4},
            0.09,
            {
                Article: 0.4,
            },
            0,
        ),  # default threshold is taken from settings with the key 'default' == 0
    ],
)
def test_init_thresholds(
    thresholds_settings,
    default_threshold,
    expected_thresholds,
    expected_default_thresholds,
):
    mock_crawler = MagicMock(spec=["spider", "stats"])
    initial_thresholds_settings = deepcopy(thresholds_settings)
    mock_crawler.spider.settings.get.return_value = thresholds_settings
    DropLowProbabilityItemPipeline.DEFAULT_THRESHOLD = default_threshold
    pipeline = DropLowProbabilityItemPipeline(mock_crawler)
    assert pipeline.thresholds_for_item == expected_thresholds
    assert pipeline.default_threshold == expected_default_thresholds
    assert initial_thresholds_settings == thresholds_settings


@pytest.mark.parametrize(
    "item, thresholds_settings, default_threshold, expected_threshold",
    [
        (
            Article(url="http://example.com"),
            {Article: 0.3, Product: 0.4},
            0.01,
            0.3,
        ),
        (
            Product(url="http://example.com"),
            {Article: 0.3, Product: 0.4},
            0.01,
            0.4,
        ),
        (Article(url="http://example.com"), {}, 0.01, 0.01),
        (Article(url="http://example.com"), {Product: 0.2}, 0.01, 0.01),
    ],
)
def test_get_threshold_for_item(
    thresholds_settings, item, default_threshold, expected_threshold
):
    mock_crawler = MagicMock(spec=["spider", "stats"])
    mock_crawler.spider.settings.get.return_value = thresholds_settings
    pipeline = DropLowProbabilityItemPipeline(mock_crawler)
    pipeline.default_threshold = default_threshold

    threshold = pipeline.get_threshold_for_item(item, mock_crawler.spider)

    assert threshold == expected_threshold


@pytest.mark.parametrize(
    "items, item_proba, threshold, expected_stats_calls, expected_return",
    [
        (
            [
                MagicMock(spec=Product(url="http://example.com")),
                MagicMock(spec=Article(url="http://example.com")),
            ],
            None,
            0.1,
            [
                ("drop_low_probability_item/processed", 2),
                ("drop_low_probability_item/processed/Product", 1),
                ("drop_low_probability_item/processed/Article", 1),
                ("drop_low_probability_item/kept", 2),
                ("drop_low_probability_item/kept/Product", 1),
                ("drop_low_probability_item/kept/Article", 1),
            ],
            True,
        ),
        (
            [
                MagicMock(spec=Product(url="http://example.com")),
                MagicMock(spec=Article(url="http://example.com")),
            ],
            0.5,
            0.1,
            [
                ("drop_low_probability_item/processed", 2),
                ("drop_low_probability_item/processed/Product", 1),
                ("drop_low_probability_item/processed/Article", 1),
                ("drop_low_probability_item/kept", 2),
                ("drop_low_probability_item/kept/Product", 1),
                ("drop_low_probability_item/kept/Article", 1),
            ],
            True,
        ),
        (
            [
                MagicMock(spec=Product(url="http://example.com")),
                MagicMock(spec=Article(url="http://example.com")),
            ],
            0.01,
            0.1,
            [
                ("drop_low_probability_item/processed", 2),
                ("drop_low_probability_item/processed/Product", 1),
                ("drop_low_probability_item/processed/Article", 1),
                ("drop_low_probability_item/dropped", 2),
                ("drop_low_probability_item/dropped/Product", 1),
                ("drop_low_probability_item/dropped/Article", 1),
            ],
            None,
        ),
    ],
)
def test_process_item(
    items, item_proba, threshold, expected_stats_calls, expected_return
):
    mock_crawler = MagicMock(spec=["spider", "stats"])

    pipeline = DropLowProbabilityItemPipeline(mock_crawler)
    with patch.object(
        DropLowProbabilityItemPipeline, "get_threshold_for_item"
    ) as mock_get_threshold_for_item:
        mock_get_threshold_for_item.return_value = threshold

        for item in items:
            item.get_probability.return_value = item_proba
            try:
                returned_item = pipeline.process_item(item, mock_crawler.spider)
            except scrapy.exceptions.DropItem as e:
                assert (
                    f"This item is dropped since the probability ({item_proba}) "
                    f"is below the threshold ({threshold}):\n{item!r}"
                ) in str(e)
            else:
                assert returned_item == item

        for call, count in expected_stats_calls:
            calls = [
                args
                for args, kwargs in mock_crawler.stats.inc_value.call_args_list
                if args[0] == call
            ]
            assert len(calls) == count


@pytest.mark.parametrize(
    "item, expected_name",
    [
        (
            Article(url="http://example.com"),
            "Article",
        ),
        (
            Product(url="http://example.com"),
            "Product",
        ),
        (
            ProductNavigation(url="http://example.com"),
            "ProductNavigation",
        ),
        (
            ArticleListPage(response=MagicMock()),
            "ArticleListPage",
        ),
    ],
)
def test_get_item_name(item, expected_name):
    assert (
        DropLowProbabilityItemPipeline.get_item_name(MagicMock(), item) == expected_name
    )



import attrs

from zyte_common_items import Item
from zyte_common_items.util import url_to_str


@attrs.define
class _Media(Item):
    #: URL.
    #:
    #: When multiple URLs exist for a given media element, pointing to
    #: different-quality versions, the highest-quality URL should be used.
    #:
    #: `Data URIs`_ are not allowed in this attribute.
    #:
    #: .. _Data URIs: https://en.wikipedia.org/wiki/Data_URI_scheme
    url: str = attrs.field(converter=url_to_str)


@attrs.define
class Image(_Media):
    """Image.

    See for example
    :class:`Product.images <zyte_common_items.Product.images>` and
    :class:`Product.mainImage <zyte_common_items.Product.mainImage>`.
    """


@attrs.define
class Audio(_Media):
    """Audio.

    See :class:`Article.audios <zyte_common_items.Article.audios>`.
    """


@attrs.define
class Video(_Media):
    """Video.

    See :class:`Article.videos <zyte_common_items.Article.videos>`.
    """



from copy import deepcopy
from warnings import warn

from zyte_common_items import ae


class AEPipeline:
    """Replace standard items with matching items with the old Zyte Automatic
    Extraction schema.

    This item pipeline is intended to help in the `migration from Zyte
    Automatic Extraction to Zyte API automatic extraction
    <https://docs.zyte.com/zyte-api/migration/zyte/autoextract.html>`_.

    In the simplest scenarios, it can be added to the ``ITEM_PIPELINES``
    setting in migrated code to ensure that the schema of output items matches
    the old schema.

    In scenarios where page object classes were being used to fix, extend or
    customize extraction, it is recommended to migrate page object classes to
    the new schemas, or move page object class code to the corresponding spider
    callback.

    If you have callbacks with custom code based on the old schema, you can
    either migrate that code, and ideally move it to a page object class, or
    use zyte_common_items.ae.downgrade at the beginning of the callback, e.g.:

    .. code-block:: python

        from zyte_common_items import ae

        ...


        def parse_product(self, response: DummyResponse, product: Product):
            product = ae.downgrade(product)
            ...
    """

    def __init__(self):
        warn(
            (
                "The zyte_common_items.pipelines.AEPipeline item pipeline has "
                "been implemented temporarily to help speed up migrating from "
                "Zyte Automatic Extraction to Zyte API automatic extraction "
                "(https://docs.zyte.com/zyte-api/migration/zyte/autoextract.html). "
                "However, this item pipeline will eventually be removed. "
                "Please, update your code not to depend on this item pipeline "
                "anymore."
            ),
            DeprecationWarning,
            stacklevel=2,
        )

    def process_item(self, item, spider):
        return ae.downgrade(item)


class DropLowProbabilityItemPipeline:
    """This pipeline drops an item if its probability, defined in the settings,
    is less than the specified threshold.

    By default, 0.1 threshold is used, i.e. items with probabillity < 0.1 are dropped.

    You can customize the thresholds by using the ITEM_PROBABILITY_THRESHOLDS setting that offers
    greater flexibility, allowing you to define thresholds for each Item class separately or
    set a default threshold for all other item classes.

    Thresholds for Item classes can be defined using either the path to the Item class or
    directly using the Item classes themselves.

    The example of using ITEM_PROBABILITY_THRESHOLDS:

    .. code-block:: python

        from zyte_common_items import Article

        ITEM_PROBABILITY_THRESHOLDS = {
            Article: 0.2,
            "zyte_common_items.Product": 0.3,
            "default": 0.15,
        }
    """

    DEFAULT_THRESHOLD = 0.1

    def __init__(self, crawler):
        self.stats = crawler.stats
        self.thresholds_for_item = {}
        self.default_threshold = None
        self.init_thresholds(crawler.spider)

    @classmethod
    def from_crawler(cls, crawler):
        return cls(crawler)

    def init_thresholds(self, spider):
        from scrapy.utils.misc import load_object

        thresholds_settings = deepcopy(
            spider.settings.get("ITEM_PROBABILITY_THRESHOLDS", {})
        )

        self.default_threshold = thresholds_settings.pop(
            "default", self.DEFAULT_THRESHOLD
        )

        for item, threshold in thresholds_settings.items():
            item_type = load_object(item) if isinstance(item, str) else item
            self.thresholds_for_item[item_type] = threshold

    def get_threshold_for_item(self, item, spider):
        return self.thresholds_for_item.get(type(item), self.default_threshold)

    def get_item_name(self, item):
        return item.__class__.__name__

    def process_item(self, item, spider):
        from scrapy.exceptions import DropItem

        item_name = self.get_item_name(item)
        item_proba = item.get_probability()
        threshold = self.get_threshold_for_item(item, spider)

        self.stats.inc_value("drop_low_probability_item/processed")
        self.stats.inc_value(f"drop_low_probability_item/processed/{item_name}")

        if item_proba is None or item_proba >= threshold:
            self.stats.inc_value("drop_low_probability_item/kept")
            self.stats.inc_value(f"drop_low_probability_item/kept/{item_name}")
            return item

        self.stats.inc_value("drop_low_probability_item/dropped")
        self.stats.inc_value(f"drop_low_probability_item/dropped/{item_name}")

        raise DropItem(
            f"This item is dropped since the probability ({item_proba}) "
            f"is below the threshold ({threshold}):\n{item!r}"
        )



from typing import List, Optional

import attrs

from zyte_common_items.base import Item
from zyte_common_items.components import (
    AdditionalProperty,
    Address,
    Breadcrumb,
    DetailsMetadata,
    Image,
    RealEstateArea,
)
from zyte_common_items.converters import to_metadata_optional, url_to_str


@attrs.define(kw_only=True)
class RealEstateMetadata(DetailsMetadata):
    """Metadata class for :data:`zyte_common_items.RealEstate.metadata`."""


@attrs.define(slots=True, kw_only=True)
class RealEstate(Item):
    """Real state offer, typically seen on real estate offer aggregator
    websites.

    :attr:`url` is the only required attribute.
    """

    #: The url of the final response, after any redirects.
    url: str = attrs.field(converter=url_to_str)

    #: Webpage `breadcrumb trail`_.
    #:
    #: .. _Breadcrumb trail: https://en.wikipedia.org/wiki/Breadcrumb_navigation
    breadcrumbs: Optional[List[Breadcrumb]] = None

    #: The identifier of the real estate, usually assigned by the seller and unique within a website, similar to product SKU.
    realEstateId: Optional[str] = None

    #: The name of the real estate.
    name: Optional[str] = None

    #: Publication date of the real estate offer.
    #:
    #: Format: ISO 8601 format: "YYYY-MM-DDThh:mm:ssZ"
    #:
    #: With timezone, if available.
    datePublished: Optional[str] = None

    #: Same date as datePublished, but before parsing/normalization, i.e. as it appears on the website.
    datePublishedRaw: Optional[str] = None

    #: The description of the real estate.
    #:
    #: Format:
    #:
    #: - trimmed (no whitespace at the beginning or the end of the description string),
    #:
    #: - line breaks included,
    #:
    #: - no length limit,
    #:
    #: - no normalization of Unicode characters,
    #:
    #: - no concatenation of description from different parts of the page.
    description: Optional[str] = None

    #: The details of the main image of the real estate.
    mainImage: Optional[Image] = None

    #: A list of URL values of all images of the real estate.
    images: Optional[List[Image]] = None

    #: The details of the address of the real estate.
    address: Optional[Address] = None

    #: Real estate area details.
    area: Optional[RealEstateArea] = None

    #: The total number of bathrooms in the real estate.
    numberOfBathroomsTotal: Optional[int] = None

    #: The number of full bathrooms in the real estate.
    numberOfFullBathrooms: Optional[int] = None

    #: The number of partial bathrooms in the real estate.
    numberOfPartialBathrooms: Optional[int] = None

    #: The number of bedrooms in the real estate.
    numberOfBedrooms: Optional[int] = None

    #: The number of rooms (excluding bathrooms and closets) of the real estate.
    numberOfRooms: Optional[int] = None

    #: Type of a trade action: buying or renting.
    tradeType: Optional[str] = None

    #: The offer price of the real estate.
    price: Optional[str] = None

    #: The rental period to which the rental price applies, only available in case of rental. Usually weekly, monthly, quarterly, yearly.
    rentalPeriod: Optional[str] = None

    #: Currency associated with the price, as appears on the page (no post-processing).
    currencyRaw: Optional[str] = None

    #: The currency of the price, in 3-letter ISO 4217 format.
    currency: Optional[str] = None

    #: A name-value pair field holding information pertaining to specific features. Usually in a form of a specification table or freeform specification list.
    additionalProperties: Optional[List[AdditionalProperty]] = None

    #: Type of the property, e.g. flat, house, land.
    propertyType: Optional[str] = None

    #: The year the real estate was built.
    yearBuilt: Optional[int] = None

    #: The URL of the virtual tour of the real estate.
    virtualTourUrl: Optional[str] = None

    #: Contains metadata about the data extraction process.
    metadata: Optional[RealEstateMetadata] = attrs.field(
        default=None, converter=to_metadata_optional(RealEstateMetadata), kw_only=True  # type: ignore[misc]
    )



from typing import List, Optional

import attrs
from web_poet import Returns

from zyte_common_items.components import Reactions, SocialMediaPostAuthor, Url
from zyte_common_items.fields import auto_field
from zyte_common_items.items import SocialMediaPost, SocialMediaPostMetadata

from .base import BasePage, Page
from .mixins import HasMetadata


class BaseSocialMediaPostPage(
    BasePage, Returns[SocialMediaPost], HasMetadata[SocialMediaPostMetadata]
):
    pass


class SocialMediaPostPage(
    Page, Returns[SocialMediaPost], HasMetadata[SocialMediaPostMetadata]
):
    pass


@attrs.define
class AutoSocialMediaPostPage(BaseSocialMediaPostPage):
    social_media_post: SocialMediaPost

    @auto_field
    def url(self) -> Optional[str]:
        return self.social_media_post.url

    @auto_field
    def postId(self) -> Optional[str]:
        return self.social_media_post.postId

    @auto_field
    def reactions(self) -> Optional[Reactions]:
        return self.social_media_post.reactions

    @auto_field
    def text(self) -> Optional[str]:
        return self.social_media_post.text

    @auto_field
    def datePublished(self) -> Optional[str]:
        return self.social_media_post.datePublished

    @auto_field
    def hashtags(self) -> Optional[List[str]]:
        return self.social_media_post.hashtags

    @auto_field
    def mediaUrls(self) -> Optional[List[Url]]:
        return self.social_media_post.mediaUrls

    @auto_field
    def author(self) -> Optional[SocialMediaPostAuthor]:
        return self.social_media_post.author

    @auto_field
    def metadata(self) -> Optional[SocialMediaPostMetadata]:
        return self.social_media_post.metadata



import warnings


class RequestListCaster:
    """Deprecated."""

    def __init__(self, target):
        warnings.warn(
            "RequestListCaster is deprecated. Please use"
            "zyte_common_items.converters.to_probability_request_list",
            DeprecationWarning,
            stacklevel=2,
        )
        self._target = target

    def __call__(self, value):
        return [item.cast(self._target) for item in value]



import html
from typing import Any, Generic, Optional, Type, Union

import html_text
from clear_html import cleaned_node_to_text
from lxml.html import HtmlElement
from price_parser import Price
from web_poet.fields import FieldsMixin, field
from web_poet.utils import ensure_awaitable, get_generic_param

from zyte_common_items.components import MetadataT


class HasMetadata(Generic[MetadataT]):
    """Inherit from this generic mixin to set the metadata class used by a page
    class."""

    @property
    def metadata_cls(self) -> Optional[Type[MetadataT]]:
        """Metadata class."""
        return _get_metadata_class(type(self))


def _get_metadata_class(cls: type) -> Optional[Type[MetadataT]]:
    return get_generic_param(cls, HasMetadata)


class PriceMixin(FieldsMixin):
    """Provides price-related field implementations."""

    _parsed_price: Optional[Price] = None

    async def _get_parsed_price(self) -> Optional[Price]:
        if self._parsed_price is None:
            # the price field wasn't executed or doesn't write _parsed_price
            price = getattr(self, "price", None)
            price = await ensure_awaitable(price)
            if self._parsed_price is None:
                # the price field doesn't write _parsed_price (or doesn't exist)
                self._parsed_price = Price(
                    amount=None, currency=None, amount_text=price
                )
        return self._parsed_price

    @field
    def currency(self) -> Optional[str]:
        return getattr(self, "CURRENCY", None)

    @field
    async def currencyRaw(self) -> Optional[str]:
        parsed_price = await self._get_parsed_price()
        if parsed_price:
            return parsed_price.currency
        return None


class DescriptionMixin(FieldsMixin):
    """Provides description and descriptionHtml field implementations."""

    UNSET = object()

    _descriptionHtml_node: Any = UNSET
    _description_node: Any = UNSET
    _description_str: Any = UNSET

    _description_default = False
    _descriptionHtml_default = False

    @staticmethod
    def wrap_description_into_html(description: str) -> str:
        r"""Convert plain text into an article HTML.

        The format tries to match clear_html.cleaned_node_to_html().

        >>> DescriptionMixin.wrap_description_into_html('')
        '<article>\n\n</article>'
        >>> DescriptionMixin.wrap_description_into_html('foo')
        '<article>\n\n<p>foo</p>\n\n</article>'
        >>> DescriptionMixin.wrap_description_into_html('foo\nbar')
        '<article>\n\n<p>foo</p>\n\n<p>bar</p>\n\n</article>'
        >>> DescriptionMixin.wrap_description_into_html('foo\n\nbar')
        '<article>\n\n<p>foo</p>\n\n<p>bar</p>\n\n</article>'
        >>> DescriptionMixin.wrap_description_into_html('\nfoo\n\nbar\n')
        '<article>\n\n<p>foo</p>\n\n<p>bar</p>\n\n</article>'
        >>> DescriptionMixin.wrap_description_into_html('foo\nbar\n\nbaz\n')
        '<article>\n\n<p>foo</p>\n\n<p>bar</p>\n\n<p>baz</p>\n\n</article>'
        >>> DescriptionMixin.wrap_description_into_html('2>1')
        '<article>\n\n<p>2&gt;1</p>\n\n</article>'
        >>> DescriptionMixin.wrap_description_into_html('<p>')
        '<article>\n\n<p>&lt;p&gt;</p>\n\n</article>'
        >>> DescriptionMixin.wrap_description_into_html('&lt;p&gt;')
        '<article>\n\n<p>&amp;lt;p&amp;gt;</p>\n\n</article>'
        """
        paras_wrapped = [
            f"\n<p>{html.escape(para)}</p>\n"
            for para in description.split("\n")
            if para
        ]
        return f"<article>\n{''.join(paras_wrapped)}\n</article>"

    async def _get_description(self) -> Optional[str]:
        if self._description_default:
            return None
        if self._description_str == self.UNSET:
            description = await ensure_awaitable(self.description)
            if self._description_str == self.UNSET:
                # the description field doesn't write _description_str
                self._description_str = description
        return self._description_str

    async def _get_description_html(self) -> Optional[HtmlElement]:
        if self._descriptionHtml_default:
            return None
        if self._descriptionHtml_node == self.UNSET:
            descriptionHtml = await ensure_awaitable(self.descriptionHtml)
            if self._descriptionHtml_node == self.UNSET:
                # the descriptionHtml field doesn't write _descriptionHtml_node
                self._descriptionHtml_node = descriptionHtml
        return self._descriptionHtml_node

    @field
    async def description(self) -> Optional[str]:
        self._description_default = True
        description_html = await self._get_description_html()
        if isinstance(description_html, HtmlElement):
            return cleaned_node_to_text(description_html)
        if isinstance(description_html, str):
            return html_text.extract_text(description_html)
        return None

    @field
    async def descriptionHtml(self) -> Union[HtmlElement, str, None]:
        self._descriptionHtml_default = True
        description = await self._get_description()
        if self._description_node not in {self.UNSET, None}:
            # we can use the element provided by the description field
            return self._description_node
        if isinstance(description, str):
            return self.wrap_description_into_html(description)
        return None



from typing import Type

import pytest
from lxml.html import fromstring
from parsel import Selector, SelectorList
from price_parser import Price
from web_poet import HttpResponse, field
from zyte_parsers import Breadcrumb as zp_Breadcrumb
from zyte_parsers import Gtin as zp_Gtin
from zyte_parsers import extract_breadcrumbs

from zyte_common_items import (
    AggregateRating,
    BasePage,
    Brand,
    Breadcrumb,
    Gtin,
    Image,
    ProductPage,
)
from zyte_common_items.components.metadata import Metadata
from zyte_common_items.items.product import ProductMetadata
from zyte_common_items.processors import (
    _format_price,
    brand_processor,
    breadcrumbs_processor,
    gtin_processor,
    images_processor,
    metadata_processor,
    price_processor,
    rating_processor,
)

base_url = "http://www.example.com/blog/"

breadcrumbs_html = """
<div class="pagesbar">
  <ul>
    <li><a href="/blog/">Home</a></li>
    <li><a href="/blog/about/">About</a></li>
  </ul>
</div>
"""

breadcrumbs_expected = [
    Breadcrumb(name="Home", url="http://www.example.com/blog/"),
    Breadcrumb(name="About", url="http://www.example.com/blog/about/"),
]


@pytest.mark.parametrize(
    "input_value,expected_value",
    [
        (None, None),
        ([], []),
        ("foo", "foo"),
        (Selector(text="<html></html>"), None),
        (SelectorList([]), None),
        (breadcrumbs_expected, breadcrumbs_expected),
        (fromstring(breadcrumbs_html), breadcrumbs_expected),
        (Selector(text=breadcrumbs_html), breadcrumbs_expected),
        (
            [
                zp_Breadcrumb(name="Home", url="http://www.example.com/blog/"),
                zp_Breadcrumb(name="About", url="http://www.example.com/blog/about/"),
            ],
            breadcrumbs_expected,
        ),
    ],
)
def test_breadcrumbs(input_value, expected_value):
    class BreadcrumbsPage(BasePage):
        @field(out=[breadcrumbs_processor])
        def breadcrumbs(self):
            return input_value

    page = BreadcrumbsPage(base_url)  # type: ignore[arg-type]
    assert page.breadcrumbs == expected_value


def test_breadcrumbs_page():
    class MyProductPage(ProductPage):
        @field
        def breadcrumbs(self):
            return self.css(".pagesbar")

    response = HttpResponse(
        url="http://www.example.com/blog/",
        body=f"<html><body>{breadcrumbs_html}</body></html>".encode(),
    )
    page = MyProductPage(response=response)
    assert page.breadcrumbs == breadcrumbs_expected


def test_breadcrumbs_page_zyte_parsers():
    class MyProductPage(ProductPage):
        @field
        def breadcrumbs(self):
            return extract_breadcrumbs(self.css(".pagesbar")[0], base_url=self.url)

    response = HttpResponse(
        url="http://www.example.com/blog/",
        body=f"<html><body>{breadcrumbs_html}</body></html>".encode(),
    )
    page = MyProductPage(response=response)
    assert page.breadcrumbs == breadcrumbs_expected


def test_breadcrumbs_base_url():
    class MyProductPage(ProductPage):
        @field
        def breadcrumbs(self):
            return self.css(".pagesbar")

    breadcrumbs_html_base = """
    <html>
    <head>
        <base href="http://www.example.com/blog/">
    </head>
    <body>
    <div class="pagesbar">
      <ul>
        <li><a href=".">Home</a></li>
        <li><a href="about/">About</a></li>
      </ul>
    </div>
    </body>
    </html>
    """

    response = HttpResponse(
        url="http://www.example.com/",
        body=breadcrumbs_html_base.encode(),
    )
    page = MyProductPage(response=response)
    assert page.breadcrumbs == breadcrumbs_expected


@pytest.mark.parametrize(
    "input_value,expected_value",
    [
        (None, None),
        ("", None),
        (" ", None),
        ("foo", Brand(name="foo")),
        (" foo ", Brand(name="foo")),
        (Selector(text="<html></html>"), None),
        (SelectorList([]), None),
        (fromstring("<p>foo</p>"), Brand(name="foo")),
        (fromstring("<img alt='foo'>"), Brand(name="foo")),
        (fromstring("<p><img alt='foo'></p>"), Brand(name="foo")),
        (fromstring("<p><p><img alt='foo'></p></p>"), Brand(name="foo")),
        (Selector(text="<p>foo</p>"), Brand(name="foo")),
        (SelectorList([Selector(text="<p>foo</p>")]), Brand(name="foo")),
    ],
)
def test_brand(input_value, expected_value):
    class BrandPage(BasePage):
        @field(out=[brand_processor])
        def brand(self):
            return input_value

    page = BrandPage("http://www.example.com/")  # type: ignore[arg-type]
    assert page.brand == expected_value


def test_brand_page():
    class MyProductPage(ProductPage):
        @field
        def brand(self):
            return self.css("body")

    response = HttpResponse(
        url="http://www.example.com/",
        body="<html><body><img alt='foo'></body></html>".encode(),
    )
    page = MyProductPage(response=response)
    assert page.brand == Brand(name="foo")


@pytest.mark.parametrize(
    "input_value,expected_value",
    [
        ("", None),
        ("NaN", None),
        ("Infinity", None),
        ("1", "1.00"),
        ("1.2", "1.20"),
        ("1.23", "1.23"),
        # Fails due to . being considered a thousands separator in this case.
        # ("1.234", "1.234"),
        ("1.2345", "1.2345"),
    ],
)
def test_format_price(input_value, expected_value):
    assert _format_price(Price.fromstring(input_value)) == expected_value


gtin_str = "978-1-933624-34-1"
gtin_html = f"<span>{gtin_str}</span>"
gtin_expected = [Gtin("isbn13", "9781933624341")]


@pytest.mark.parametrize(
    "input_value,expected_value",
    [
        (None, None),
        ([], None),
        ("foo", None),
        (Selector(text="<html></html>"), None),
        (SelectorList([]), None),
        (gtin_expected, gtin_expected),
        (fromstring(gtin_html), gtin_expected),
        (Selector(text=gtin_html), gtin_expected),
        (
            [
                zp_Gtin("isbn13", "9781933624341"),
                zp_Gtin("isbn13", "9780525555360"),
            ],
            [
                Gtin("isbn13", "9781933624341"),
                Gtin("isbn13", "9780525555360"),
            ],
        ),
        (gtin_str, gtin_expected),
        (
            [
                "978-1-933624-34-1",
                "97-805-25555-360",
            ],
            [
                Gtin("isbn13", "9781933624341"),
                Gtin("isbn13", "9780525555360"),
            ],
        ),
    ],
)
def test_gtin(input_value, expected_value):
    class GtinPage(BasePage):
        @field(out=[gtin_processor])
        def gtin(self):
            return input_value

    page = GtinPage(base_url)  # type: ignore[arg-type]
    assert page.gtin == expected_value


@pytest.mark.parametrize(
    "input_value,expected_value",
    [
        (None, None),
        ([], []),
        ("foo", "foo"),
        (Selector(text="<html></html>"), None),
        (SelectorList([]), None),
        (Selector(text="<html>3.8</html>"), AggregateRating(ratingValue=3.8)),
        (
            Selector(text="<html>3.8 out of 10</html>"),
            AggregateRating(ratingValue=3.8, bestRating=10.0),
        ),
        (
            Selector(text="<html>3.8 (7 reviews)</html>"),
            AggregateRating(ratingValue=3.8, reviewCount=7),
        ),
        # provided as a separate xfail test
        # (
        #     Selector(text="<html>3.8 out of 10 (5 reviews)</html>"),
        #     AggregateRating(ratingValue=3.8, reviewCount=5),
        # ),
        (
            AggregateRating(ratingValue=3.8, bestRating=5.0, reviewCount=3),
            AggregateRating(ratingValue=3.8, bestRating=5.0, reviewCount=3),
        ),
        (
            {"ratingValue": 3.8, "bestRating": 5.0, "reviewCount": 3},
            AggregateRating(ratingValue=3.8, bestRating=5.0, reviewCount=3),
        ),
        (
            {"ratingValue": 3.8, "reviewCount": 3},
            AggregateRating(ratingValue=3.8, reviewCount=3),
        ),
        (
            {"ratingValue": 3.8},
            AggregateRating(ratingValue=3.8),
        ),
        (
            {"ratingValue": 3.8, "reviewCount": "3"},
            AggregateRating(ratingValue=3.8, reviewCount=3),
        ),
        (
            {"ratingValue": 3.8, "bestRating": 10, "reviewCount": 3},
            AggregateRating(ratingValue=3.8, bestRating=10.0, reviewCount=3),
        ),
        (
            {
                "ratingValue": Selector(text="<html>3.8 out of 10</html>"),
                "reviewCount": 3,
            },
            AggregateRating(ratingValue=3.8, bestRating=10.0, reviewCount=3),
        ),
        (
            {
                "ratingValue": Selector(text="<html>3.8 out of 10</html>"),
                "bestRating": 5.0,
                "reviewCount": 3,
            },
            AggregateRating(ratingValue=3.8, bestRating=5.0, reviewCount=3),
        ),
        (
            {
                "ratingValue": Selector(text="<html>3.8 out of 10</html>"),
                "reviewCount": Selector(text="<html>3 reviews</html>"),
            },
            AggregateRating(ratingValue=3.8, bestRating=10.0, reviewCount=3),
        ),
    ],
)
def test_rating(input_value, expected_value):
    class RatingPage(BasePage):
        @field(out=[rating_processor])
        def aggregateRating(self):
            return input_value

    page = RatingPage(base_url)  # type: ignore[arg-type]
    assert page.aggregateRating == expected_value


@pytest.mark.xfail(
    reason="When more than 2 numbers are found bestRating is not extracted"
)
def test_rating_3_values():
    class RatingPage(BasePage):
        @field(out=[rating_processor])
        def aggregateRating(self):
            return Selector(text="<html>3.8 out of 10 (5 reviews)</html>")

    page = RatingPage(base_url)  # type: ignore[arg-type]
    assert page.aggregateRating == AggregateRating(
        ratingValue=3.8, bestRating=10, reviewCount=5
    )


@pytest.mark.parametrize(
    "input_value,expected_value",
    [
        (None, None),
        ([], []),
        ("https://www.url.com/img.jpg", [Image(url="https://www.url.com/img.jpg")]),
        (
            [
                Image("https://www.url.com/img1.jpg"),
                Image("https://www.url.com/img2.jpg"),
            ],
            [
                Image("https://www.url.com/img1.jpg"),
                Image("https://www.url.com/img2.jpg"),
            ],
        ),
        (
            ["https://www.url.com/img1.jpg", "https://www.url.com/img2.jpg"],
            [
                Image("https://www.url.com/img1.jpg"),
                Image("https://www.url.com/img2.jpg"),
            ],
        ),
        (
            [
                {"url": "https://www.url.com/img1.jpg"},
                {"url": "https://www.url.com/img2.jpg"},
            ],
            [
                Image("https://www.url.com/img1.jpg"),
                Image("https://www.url.com/img2.jpg"),
            ],
        ),
    ],
)
def test_images(input_value, expected_value):
    class ImagesPage(BasePage):
        @field(out=[images_processor])
        def images(self):
            return input_value

    page = ImagesPage(base_url)  # type: ignore[arg-type]
    assert page.images == expected_value


def test_images_page():
    class MyProductPage(ProductPage):
        @field
        def images(self):
            return self.css("img::attr(href)").getall()

    response = HttpResponse(
        url="http://www.example.com/",
        body="<html><body><img href='https://www.url.com/img.jpg'></body></html>".encode(),
    )
    page = MyProductPage(response=response)
    assert page.images == [Image(url="https://www.url.com/img.jpg")]


@pytest.mark.parametrize(
    "input_value,expected_value",
    [
        (100, "100.00"),
        (None, None),
        ([], []),
        ({}, {}),
        (22.9, "22.90"),
        (22.0, "22.00"),
        ("22.9", "22.9"),
        ("Do not apply to strings...", "Do not apply to strings..."),
    ],
)
def test_prices(input_value, expected_value):
    class PricePage(BasePage):
        @field(out=[price_processor])
        def price(self):
            return input_value

    page = PricePage(base_url)  # type: ignore[arg-type]
    assert page.price == expected_value


@pytest.mark.parametrize(
    "input_value,BasePage,expected_value",
    [
        (None, ProductPage, None),
        (Metadata(), ProductPage, ProductMetadata()),
    ],
)
def test_metadata(input_value, BasePage: Type, expected_value):
    class CustomPage(BasePage):
        @field(out=[metadata_processor])
        def metadata(self):
            return input_value

    page = CustomPage(base_url)  # type: ignore[arg-type]
    assert page.metadata == expected_value



from typing import Optional

import attrs

from zyte_common_items.base import Item
from zyte_common_items.util import url_to_str


@attrs.define(kw_only=True)
class Breadcrumb(Item):
    """A breadcrumb from the `breadcrumb trail`_ of a webpage.

    See :attr:`Product.breadcrumbs
    <zyte_common_items.Product.breadcrumbs>`.

    .. _breadcrumb trail: https://en.wikipedia.org/wiki/Breadcrumb_navigation
    """

    #: Displayed name.
    name: Optional[str] = None

    #: Target URL.
    url: Optional[str] = attrs.field(
        default=None, converter=attrs.converters.optional(url_to_str), kw_only=True
    )



# flake8: noqa
from .article import ArticlePage, AutoArticlePage, BaseArticlePage
from .article_list import ArticleListPage, AutoArticleListPage, BaseArticleListPage
from .article_navigation import (
    ArticleNavigationPage,
    AutoArticleNavigationPage,
    BaseArticleNavigationPage,
)
from .base import BasePage, Page
from .business_place import (
    AutoBusinessPlacePage,
    BaseBusinessPlacePage,
    BusinessPlacePage,
)
from .job_posting import AutoJobPostingPage, BaseJobPostingPage, JobPostingPage
from .mixins import DescriptionMixin, HasMetadata, MetadataT, PriceMixin
from .product import AutoProductPage, BaseProductPage, ProductPage
from .product_list import AutoProductListPage, BaseProductListPage, ProductListPage
from .product_navigation import (
    AutoProductNavigationPage,
    BaseProductNavigationPage,
    ProductNavigationPage,
)
from .real_estate import AutoRealEstatePage, BaseRealEstatePage, RealEstatePage
from .search_request_template import SearchRequestTemplatePage
from .social_media_post import (
    AutoSocialMediaPostPage,
    BaseSocialMediaPostPage,
    SocialMediaPostPage,
)



# flake8: noqa
from ._compat import request_list_processor
from .additional_property import AdditionalProperty
from .address import Address
from .author import Author
from .brand import Brand
from .breadcrumbs import Breadcrumb
from .business_place import Amenity, OpeningHoursItem, ParentPlace
from .gtin import Gtin
from .job_posting import BaseSalary, HiringOrganization, JobLocation
from .links import Link, NamedLink, Url
from .media import Audio, Image, Video
from .metadata import (
    BaseMetadata,
    DetailsMetadata,
    ListMetadata,
    Metadata,
    MetadataT,
    ProbabilityMetadata,
)
from .ratings import AggregateRating, StarRating
from .real_estate import RealEstateArea
from .request import Header, ProbabilityRequest, Request
from .social_media_post import Reactions, SocialMediaPostAuthor



"""This test file is ignored by mypy since we're using pytest-mypy-testing to
assert the mypy errors that should be raised.

Despite some mypy errors, some fields will still attempt proper value conversions
to the proper type. If the given value cannot be converted, a ValueError is raised.
See ``tests/test_conversion.py`` for the conversion checks.

Assertions in mypy are done via these line comments:
    # N: <msg> - assert a mypy note message
    # W: <msg> - assert a mypy warning message
    # E: <msg> - assert a mypy error message
    # R: <msg> - assert a mypy note message Revealed type is '<msg>'.

NOTE:
    Pytest's parametrization doesn't work here due to the dynamic nature of
    variables in each line. For-loops also doesn't work well since the type
    checker needs to narrow down on the type within the test function's scope.
"""

import pytest
from web_poet import HttpResponse

from zyte_common_items import (
    Breadcrumb,
    Image,
    Link,
    Product,
    ProductFromList,
    ProductList,
    ProductVariant,
)

URL = "https://www.some-page/"
RESPONSE = HttpResponse(URL, b"")


# NOTES: The following test cases below for assigning ``ResponseUrl`` instances
# are marked with "xfail" due to the following known mypy bugs:
#   - https://github.com/python/mypy/issues/10187
#   - https://github.com/python/mypy/issues/5313
#   - https://github.com/python-attrs/attrs/issues/889
# Currently, the said bugs causes mypy to raise the following error:
#   'Incompatible types in assignment (expression has type "ResponseUrl",
#   variable has type "Optional[str]")'
# FIXME: Unmark these from "xfail" once the upstream mypy issues are solved.


@pytest.mark.mypy_testing
@pytest.mark.xfail
def test_assignment_image_url():
    img = Image(url=RESPONSE.url)
    img.url = RESPONSE.url


@pytest.mark.mypy_testing
@pytest.mark.xfail
def test_assignment_breadcrumb_url():
    bc = Breadcrumb(url=RESPONSE.url)
    bc.url = RESPONSE.url


@pytest.mark.mypy_testing
@pytest.mark.xfail
def test_assignment_link_url():
    link = Link(url=RESPONSE.url)
    link.url = RESPONSE.url


@pytest.mark.mypy_testing
@pytest.mark.xfail
def test_assignment_product_url():
    product_1 = Product(url=RESPONSE.url)
    product_1.url = RESPONSE.url

    product_2 = Product(url=RESPONSE.url)
    product_2.canonicalUrl = RESPONSE.url


@pytest.mark.mypy_testing
@pytest.mark.xfail
def test_assignment_product_list_url():
    product_list_1 = ProductList(url=RESPONSE.url)
    product_list_1.url = RESPONSE.url

    product_list_2 = ProductList(url=RESPONSE.url)
    product_list_2.canonicalUrl = RESPONSE.url


@pytest.mark.mypy_testing
@pytest.mark.xfail
def test_assignment_product_from_list_url():
    product_from_list_1 = ProductFromList(url=RESPONSE.url)
    product_from_list_1.url = RESPONSE.url


@pytest.mark.mypy_testing
@pytest.mark.xfail
def test_assignment_product_variant_url():
    product_variant_1 = ProductVariant(url=RESPONSE.url)
    product_variant_1.url = RESPONSE.url

    product_variant_2 = ProductVariant(url=RESPONSE.url)
    product_variant_2.canonicalUrl = RESPONSE.url


@pytest.mark.mypy_testing
def test_assignment_image():
    img_1 = Image(url=RESPONSE.url)
    img_1.url = str(RESPONSE.url)

    img_2 = Image(url=RESPONSE.url)
    img_2.url = "https://www.example.com"


@pytest.mark.mypy_testing
def test_assignment_breadcrumb():
    bc_1 = Breadcrumb(url=RESPONSE.url)
    bc_1.url = str(RESPONSE.url)

    bc_2 = Breadcrumb(url=RESPONSE.url)
    bc_2.url = "https://www.example.com"


@pytest.mark.mypy_testing
def test_assignment_link():
    link_1 = Link(url=RESPONSE.url)
    link_1.url = str(RESPONSE.url)

    link_2 = Link(url=RESPONSE.url)
    link_2.url = "https://www.example.com"


@pytest.mark.mypy_testing
def test_assignment_product():
    product_1 = Product(url=RESPONSE.url)
    product_1.url = str(RESPONSE.url)

    product_2 = Product(url=RESPONSE.url)
    product_2.url = "https://www.example.com"

    product_3 = Product(url=RESPONSE.url)
    product_3.canonicalUrl = str(RESPONSE.url)

    product_4 = Product(url=RESPONSE.url)
    product_4.canonicalUrl = "https://www.example.com"


@pytest.mark.mypy_testing
def test_assignment_product_list():
    product_list_1 = ProductList(url=RESPONSE.url)
    product_list_1.url = str(RESPONSE.url)

    product_list_2 = ProductList(url=RESPONSE.url)
    product_list_2.url = "https://www.example.com"

    product_list_3 = ProductList(url=RESPONSE.url)
    product_list_3.canonicalUrl = str(RESPONSE.url)

    product_list_4 = ProductList(url=RESPONSE.url)
    product_list_4.canonicalUrl = "https://www.example.com"


@pytest.mark.mypy_testing
def test_assignment_product_from_list():
    product_from_list_1 = ProductFromList(url=RESPONSE.url)
    product_from_list_1.url = str(RESPONSE.url)

    product_from_list_2 = ProductFromList(url=RESPONSE.url)
    product_from_list_2.url = "https://www.example.com"


@pytest.mark.mypy_testing
def test_assignment_product_variant():
    product_variant_1 = ProductVariant(url=RESPONSE.url)
    product_variant_1.url = str(RESPONSE.url)

    product_variant_2 = ProductVariant(url=RESPONSE.url)
    product_variant_2.url = "https://www.example.com"

    product_variant_3 = ProductVariant(url=RESPONSE.url)
    product_variant_3.canonicalUrl = str(RESPONSE.url)

    product_variant_4 = ProductVariant(url=RESPONSE.url)
    product_variant_4.canonicalUrl = "https://www.example.com"


@pytest.mark.mypy_testing
def test_instantiation_image():
    with pytest.raises(ValueError):
        Image(
            url=123  # E: Argument "url" to "Image" has incompatible type "int"; expected "str | _Url"  [arg-type]
        )


@pytest.mark.mypy_testing
def test_instantiation_breadcrumb():
    with pytest.raises(ValueError):
        Breadcrumb(
            url=123  # E: Argument "url" to "Breadcrumb" has incompatible type "int"; expected "str | _Url | None"  [arg-type]
        )


@pytest.mark.mypy_testing
def test_instantiation_link():
    with pytest.raises(ValueError):
        Link(
            url=123  # E: Argument "url" to "Link" has incompatible type "int"; expected "str | _Url | None"  [arg-type]
        )


@pytest.mark.mypy_testing
def test_instantiation_product_list():
    with pytest.raises(ValueError):
        ProductList(
            url=123  # E: Argument "url" to "ProductList" has incompatible type "int"; expected "str | _Url"  [arg-type]
        )
    with pytest.raises(ValueError):
        ProductList(
            url="https://www.example.com",
            canonicalUrl=123,  # E: Argument "canonicalUrl" to "ProductList" has incompatible type "int"; expected "str | _Url | None"  [arg-type]
        )


@pytest.mark.mypy_testing
def test_instantiation_product_from_list():
    with pytest.raises(ValueError):
        ProductFromList(
            url=123  # E: Argument "url" to "ProductFromList" has incompatible type "int"; expected "str | _Url | None"  [arg-type]
        )


@pytest.mark.mypy_testing
def test_instantiation_product_variant():
    with pytest.raises(ValueError):
        ProductVariant(
            url=123  # E: Argument "url" to "ProductVariant" has incompatible type "int"; expected "str | _Url | None"  [arg-type]
        )
    with pytest.raises(ValueError):
        ProductVariant(
            url="https://www.example.com",
            canonicalUrl=123,  # E: Argument "canonicalUrl" to "ProductVariant" has incompatible type "int"; expected "str | _Url | None"  [arg-type]
        )



from base64 import b64encode
from typing import List, Optional
from urllib.parse import quote_plus

import attrs
import jinja2

from zyte_common_items.base import Item
from zyte_common_items.components import DetailsMetadata, Header, Request
from zyte_common_items.converters import to_metadata_optional

_TEMPLATE_ENVIRONMENT = jinja2.Environment()
_TEMPLATE_ENVIRONMENT.filters["quote_plus"] = quote_plus


def _render(string: str, **kwargs) -> str:
    return _TEMPLATE_ENVIRONMENT.from_string(string).render(**kwargs)


@attrs.define(kw_only=True)
class SearchRequestTemplateMetadata(DetailsMetadata):
    """Metadata class for :data:`zyte_common_items.SearchRequestTemplate.metadata`."""


@attrs.define(kw_only=True)
class SearchRequestTemplate(Item):
    """:ref:`Request template <request-templates>` to build a search
    :class:`~zyte_common_items.Request`."""

    #: :doc:`Jinja template <jinja:templates>` for :class:`Request.url
    #: <zyte_common_items.Request.url>`.
    url: str

    #: :doc:`Jinja template <jinja:templates>` for :class:`Request.method
    #: <zyte_common_items.Request.method>`.
    method: str = "GET"

    #: :doc:`Jinja template <jinja:templates>` for :class:`Request.body
    #: <zyte_common_items.Request.body>`.
    #:
    #: It must be a plain :class:`str`, not :class:`bytes` or a Base64-encoded
    #: :class:`str`. Base64-encoding is done by :meth:`request` after rendering
    #: this value as a Jinja template.
    #:
    #: Defining a non-UTF-8 body is not supported.
    body: Optional[str] = None

    #: List of :class:`Header`, for :class:`Request.headers
    #: <zyte_common_items.Request.headers>`, where every :attr:`~Header.name`
    #: and :attr:`~Header.value` is a :doc:`Jinja template <jinja:templates>`.
    #:
    #: When a header name template renders into an empty string (after
    #: stripping spacing), that header is removed from the resulting list of
    #: headers.
    headers: Optional[List[Header]] = None

    #: Data extraction process metadata.
    metadata: Optional[SearchRequestTemplateMetadata] = attrs.field(
        default=None, converter=to_metadata_optional(SearchRequestTemplateMetadata), kw_only=True  # type: ignore[misc]
    )

    def request(self, *, keyword: str) -> Request:
        """Return a :class:`~zyte_common_items.Request` to search for
        *keyword*."""
        body = _render(self.body or "", keyword=keyword).encode()

        headers = []
        for header in self.headers or []:
            name = _render(header.name, keyword=keyword).strip()
            if not name:
                continue
            value = _render(header.value, keyword=keyword)
            headers.append(Header(name=name, value=value))

        return Request(
            url=_render(self.url, keyword=keyword),
            method=_render(self.method, keyword=keyword),
            body=b64encode(body).decode() if body else None,
            headers=headers or None,
        )



from __future__ import annotations

from typing import List, Optional

import attrs

from zyte_common_items.base import Item
from zyte_common_items.components import (
    Breadcrumb,
    Image,
    Link,
    ListMetadata,
    ProbabilityMetadata,
)
from zyte_common_items.converters import (
    to_metadata_optional,
    url_to_str,
    url_to_str_optional,
)


@attrs.define(kw_only=True)
class ProductListMetadata(ListMetadata):
    """Metadata class for :data:`zyte_common_items.ProductList.metadata`."""


@attrs.define(slots=True, kw_only=True)
class ProductList(Item):
    """Product list from a product listing page of an e-commerce webpage.

    It represents, for example, a single page from a category.

    :attr:`url` is the only required attribute.
    """

    #: Webpage `breadcrumb trail`_.
    #:
    #: .. _Breadcrumb trail: https://en.wikipedia.org/wiki/Breadcrumb_navigation
    breadcrumbs: Optional[List[Breadcrumb]] = None

    #: Canonical form of the URL, as indicated by the website.
    #:
    #: See also ``url``.
    canonicalUrl: Optional[str] = attrs.field(
        default=None, converter=url_to_str_optional, kw_only=True
    )

    #: Name of the product listing as it appears on the webpage
    #: (no post-processing).
    #:
    #: For example, if the webpage is one of the pages of the Robots category,
    #: ``categoryName`` is ``'Robots'``.
    categoryName: Optional[str] = None

    #: Data extraction process metadata.
    metadata: Optional[ProductListMetadata] = attrs.field(
        default=None, converter=to_metadata_optional(ProductListMetadata), kw_only=True  # type: ignore[misc]
    )

    #: Current page number, if displayed explicitly on the list page.
    #:
    #: Numeration starts with 1.
    pageNumber: Optional[int] = None

    #: Link to the next page.
    paginationNext: Optional[Link] = None

    #: List of products.
    #:
    #: It only includes product information found in the product listing page
    #: itself. Product information that requires visiting each product URL is
    #: not meant to be covered.
    #:
    #: The order of the products reflects their position on the rendered page.
    #: Product order is top-to-bottom, and left-to-right or right-to-left
    #: depending on the webpage locale.
    products: Optional[List[ProductFromList]] = None

    #: Main URL from which the data has been extracted.
    #:
    #: See also ``canonicalUrl``.
    url: str = attrs.field(converter=url_to_str)


@attrs.define(slots=True, kw_only=True)
class ProductFromList(Item):
    """Product from a product list from a product listing page of an e-commerce
    webpage.

    See :class:`ProductList`, :class:`ProductFromListExtractor`,
    :class:`ProductFromListSelectorExtractor`.
    """

    #: Price currency `ISO 4217`_ alphabetic code (e.g. ``"USD"``).
    #:
    #: See also ``currencyRaw``.
    #:
    #: .. _ISO 4217: https://en.wikipedia.org/wiki/ISO_4217
    currency: Optional[str] = None

    #: Price currency as it appears on the webpage (no post-processing), e.g.
    #: ``"$"``.
    #:
    #: See also ``currency``.
    currencyRaw: Optional[str] = None

    #: Main product image.
    mainImage: Optional[Image] = None

    #: Data extraction process metadata.
    metadata: Optional[ProbabilityMetadata] = attrs.field(
        default=None, converter=to_metadata_optional(ProbabilityMetadata), kw_only=True  # type: ignore[misc]
    )

    #: Name as it appears on the webpage (no post-processing).
    name: Optional[str] = None

    #: Price at which the product is being offered.
    #:
    #: It is a string with the price amount, with a full stop as decimal
    #: separator, and no thousands separator or currency (see ``currency`` and
    #: ``currencyRaw``), e.g. ``"10500.99"``.
    #:
    #: If ``regularPrice`` is not ``None``, ``price`` should always be lower
    #: than ``regularPrice``.
    price: Optional[str] = None

    #: Product identifier, unique within an e-commerce website.
    #:
    #: It may come in the form of an SKU or any other identifier, a hash, or
    #: even a URL.
    productId: Optional[str] = None

    #: Price at which the product was being offered in the past, and which is
    #: presented as a reference next to the current price.
    #:
    #: It may be labeled as the original price, the list price, or the maximum
    #: retail price for which the product is sold.
    #:
    #: See ``price`` for format details.
    #:
    #: If ``regularPrice`` is not ``None``, it should always be higher than
    #: ``price``.
    regularPrice: Optional[str] = None

    #: Main URL from which the product data could be extracted.
    url: Optional[str] = attrs.field(
        default=None, converter=url_to_str_optional, kw_only=True
    )



from typing import Callable, List, Optional

from web_poet import ItemPage, field
from web_poet.fields import get_fields_dict


def auto_field(
    method=None,
    *,
    cached: bool = False,
    meta: Optional[dict] = None,
    out: Optional[List[Callable]] = None,
):
    """Decorator that works like :func:`web_poet.fields.field` but sets
    ``auto_field`` to ``True`` by default in *meta*.

    .. code-block:: python

        from zyte_common_items import AutoProductPage
        from zyte_common_items.fields import auto_field


        class ProductPage(AutoProductPage):
            @auto_field
            def name(self):
                return super().name
    """
    meta = meta or {}
    meta.setdefault("auto_field", True)
    return field(method, cached=cached, meta=meta, out=out)


def is_auto_field(cls: ItemPage, field: str):
    """Return ``True`` if the field named *field* of the *cls* page object
    class has ``auto_field`` set to ``True`` in its field metadata.

    All fields defined in :ref:`auto page object classes <auto>` meet this
    condition.
    """
    fields_dict = get_fields_dict(cls)
    field_meta = fields_dict[field].meta or {}
    return field_meta.get("auto_field", False)
